# -*- coding: utf-8 -*-

from .datawrapper import DataWrapper
from .typed import Typed
from .typecoded import Typecoded
from .listener import ColumnListener
from .ndprint import ndprint
from .shaped import Shaped
from ..utils.common import mstr, bstr, lls, exprstrs, findShape
from .dataset import GenericDataset, make_title_meta_l0
try:
    from .arraydataset_datamodel import Model
except ImportError as e:
    Model = {'metadata': {}}


from collections.abc import Sequence, Iterable
from collections import OrderedDict
from copy import copy

import logging
# create logger
logger = logging.getLogger(__name__)
# logger.debug('level %d' %  (logger.getEffectiveLevel()))

MdpInfo = Model['metadata']

# how many columns and rows to display and where to add '...'
COLUMN_WIDTH = 20
COLUMN_W_BREAK = 16
ROW_HEIGHT = 20
ROW_H_BREAK = 16


def mkline(line):
    """ make a display row from a given sequence.

    :line: the array or sequence line to display.
    """
    return list(line[:COLUMN_W_BREAK]) + ['...'] + \
        list(line[COLUMN_W_BREAK-COLUMN_WIDTH:])


class ArrayDataset(GenericDataset, Iterable, Shaped):
    """  Special dataset that contains a single Array Data object.

    mh: If omit the parameter names during instanciation, e.g. ArrayDataset(a, b, c), the assumed order is data, unit, description.
    mh:  contains a sequence which provides methods count(), index(), remove(), reverse().
    A mutable sequence would also need append(), extend(), insert(), pop() and sort().

        Parameters
        ----------
    :data: the payload data of this dataset. Default is None, Can be any iterable except for memoryview.
        Returns
        -------
    """

    def __init__(self, data=None,
                 unit=None,
                 description=None,
                 typ_=None,
                 typecode=None,
                 version=None,
                 zInfo=None,
                 alwaysMeta=True,
                 ** kwds):
        """ Initializes an ArrayDataset.

        Default ```None``` will initialize MetaData Parameters to their default values.
        If ``data`` is not None and has shape (``len`` applies), ``shape`` MDP is set to the actual ``data`` shape.
        """

        # collect MDPs from args-turned-local-variables.
        metasToBeInstalled = copy(locals())
        metasToBeInstalled.pop('__class__', None)
        metasToBeInstalled.pop('kwds', None)
        metasToBeInstalled.pop('self', None)
        metasToBeInstalled.pop('zInfo', None)

        global Model
        if zInfo is None:
            zInfo = Model

        # print('@1 zInfo', id(self.zInfo['metadata']), id(self), id(self.zInfo),
        #      self.zInfo['metadata']['version'], list(metasToBeInstalled.keys()))
        # must be the first line to initiate meta
        super().__init__(zInfo=zInfo, **metasToBeInstalled, **kwds)
        self.updateShape()

    # def getData(self):
    #     """ Optimized """
    #     return self._data

    def setData(self, data):
        """
        """

        # if issubclass(data.__class__, memoryview):
        #    d = data
        if data is None:
            d = None
        else:
            isitr = hasattr(data, '__iter__') or hasattr(data, '__getitem__')
            # and hasattr(data, '__next__')
            if not isitr:
                # dataWrapper initializes data as None
                m = 'data in ArrayDataset must be an iterator, not ' + \
                    data.__class__.__name__
                raise TypeError(m)
            else:
                d = data
        # no passive shape-updating. no
        super(ArrayDataset, self).setData(d)
        self.updateShape()

    def __setitem__(self, *args, **kwargs):
        """ sets value at key.
        """
        self.getData().__setitem__(*args, **kwargs)
        self.updateShape()

    def __getitem__(self, *args, **kwargs):
        """ returns value at key.
        """
        return self.getData().__getitem__(*args, **kwargs)

    def __delitem__(self, *args, **kwargs):
        """ removes value and its key.
        """
        self.getData().__delitem__(*args, **kwargs)

    def __iter__(self, *args, **kwargs):
        """ returns an iterator
        """
        return self.getData().__iter__(*args, **kwargs)

    def pop(self, *args, **kwargs):
        """ revomes and returns value
        """
        ret = self.getData().pop(*args, **kwargs)
        self.updateShape()
        return ret

    def append(self, *args, **kwargs):
        """ appends to data.
        """
        self.getData().append(*args, **kwargs)
        self.updateShape()

    def extend(self, *args, **kwargs):
        """ extend data.
        """
        self.getData().extend(*args, **kwargs)
        self.updateShape()

    def index(self, *args, **kwargs):
        """ returns the index of a value.
        """
        return self.getData().index(*args, **kwargs)

    def count(self, *args, **kwargs):
        """ returns size.
        """
        return self.getData().count(*args, **kwargs)

    def remove(self, *args, **kwargs):
        """ removes value at first occurrence.
        """
        self.getData().remove(*args, **kwargs)
        self.updateShape()

    def __repr__(self):
        return self.toString(level=2)

    def toString(self, level=0, param_widths=None,
                 tablefmt='grid', tablefmt1='simple', tablefmt2='simple',
                 width=0, matprint=None, trans=True,
                 center=-1, heavy=True, **kwds):
        """ matprint: an external matrix print function
        trans: print 2D matrix transposed. default is True.
        """
        if matprint is None:
            matprint = ndprint

        cn = self.__class__.__name__
        if level > 1:

            s = cn + '(' +\
                self.meta.toString(
                    level=level, heavy=heavy,
                    tablefmt=tablefmt, tablefmt1=tablefmt1, tablefmt2=tablefmt2,
                    width=width, param_widths=param_widths,
                    **kwds)
            # set wiidth=0 level=2 to inhibit \n
            vs, us, ts, ds, fs, gs, cs, ex = exprstrs(
                self, '_data', width=0, level=level)
            # '{ %s (%s) <%s>, "%s", default %s, tcode=%s}' %\
            # (vs, us, ts, ds, fs, cs)
            return '%s data= %s)' % (s, vs)

        html = 'html' in tablefmt.lower() or 'html' in tablefmt2.lower()
        br = '<br>' if html else '\n'
        if html:
            tablefmt = tablefmt2 = 'unsafehtml'

        s, last = make_title_meta_l0(self, level=level, width=width, heavy=heavy,
                                     tablefmt=tablefmt, tablefmt1=tablefmt1,
                                     tablefmt2=tablefmt2, center=center,
                                     param_widths=param_widths,
                                     html=html, excpt=['description'])
        width = len(last)-1
        if level == 0:
            if html:
                d = '<center><u>%s</u></center>\n' % 'DATA' + '<hr>'
            else:
                d = 'DATA'.center(width) + '\n' + '----'.center(width) + '\n'
        else:
            d = ''

        # limit max rows and columns of display
        ls = len(self.shape)
        if ls == 0 or \
           ls == 1 and len(self.data) <= COLUMN_WIDTH or \
           ls > 1 and len(self.data[0]) <= COLUMN_WIDTH and \
           len(self.data) <= ROW_HEIGHT:
            tdata = self.data
        elif ls == 1:
            tdata = mkline(self.data)
        else:
            # ls > 1
            if len(self.data) <= ROW_HEIGHT:
                tdata = [mkline(line) for line in self.data]
            else:
                tdata = [mkline(line) for line in self.data[:ROW_H_BREAK]]
                tdata += [['...'] * len(tdata[0])]
                tdata += [mkline(line)
                          for line in self.data[ROW_H_BREAK-ROW_HEIGHT:]]

        ds = bstr(tdata, level=level, **kwds) if matprint is None else \
            matprint(tdata, trans=False, headers=[],
                     tablefmt2='html' if html else 'plain',
                     **kwds)
        #d += lls(ds, 9000 if html else 2000)
        d += ds
        return '%s\n%s%s%s%s' % (s, d, br, last, br)

    string = toString
    txt = toString

    def __getstate__(self):
        """ Can be encoded with serializableEncoder """

        # s = OrderedDict(description=self.description, meta=self.meta, data=self.data)  # super(...).__getstate__()
        s = OrderedDict(
            _ATTR_meta=getattr(self, '_meta', None),
            _ATTR_data=getattr(self, 'data', None))

        return s
        # type=self.type,
        # unit=self.unit,
        # typecode=self.typecode,
        # version=self.version,
        # FORMATV=self.FORMATV,


class Column(ArrayDataset, ColumnListener):
    """ A Column is a the vertical cut of a table for which all cells have the same signature.

    A Column contains raw ArrayData, and optionally a description and unit.
    example::

      table = TableDataset()
      table.addColumn("Energy",Column(data=[1,2,3,4],description="desc",unit='eV'))
    """

    def __init__(self,  *args, typ_='Column', **kwds):
        super().__init__(*args, typ_=typ_, **kwds)



# -*- coding: utf-8 -*-

from .common import lls, trbk
import pwd
import os
import signal
from subprocess import (Popen,
                        PIPE,
                        TimeoutExpired,
                        )

import logging
# create logger
logger = logging.getLogger(__name__)


def run_proc(cmd, as_user, pwdir, timeout, pty=PIPE):
    """Execute a shell command and return status.

    Parameters
    ----------
    cmd : list
        A list of command line strings. e.g. from ``shelex.split("ls -t")``
    as_user : str
        execute as user. overrides env var `USER` and
        `pwd.getpwnam()`.
    pwdir : str
        overides what is in `pwd.getpwnam()`.
    timeout : float
        sets `popen` timeout. If triggers will get the processed
        killed.
    pty : int
        passed to `stdin`, `stdout`, `stderr` in `popen`.
    pid_file: str2
        Name of the file that has the `cmd` process ID. Needed to kill by grand parents.

    Returns
    -------
    dict
        If  `try` block fails when starting, a dictionary of:

    :returncode: the process return code. -9 if killed aftet timeout.
    :message: Error message from exception.

         If starting successfully, a dictionary of:

    :command: The `cmd` from input.
    :returncode: the process `returncode`. `-9` if killed aftet timeout.
    :stdout:, stderr: from `communicate()` of the process.
    :message: Announce success and why may have been killed.
    """

    try:
        # https://stackoverflow.com/a/6037494

        user_name = as_user if as_user else os.environ.get('USER')
        pw_record = pwd.getpwnam(user_name)
        user_name = pw_record.pw_name
        user_home_dir = pw_record.pw_dir
        user_uid = pw_record.pw_uid
        user_gid = pw_record.pw_gid
        env = os.environ.copy()
        env['HOME'] = user_home_dir
        env['LOGNAME'] = env['USER'] = user_name
        env['PWD'] = pwdir
        env['SystemRoot'] = ''

        executable = None

        def chusr(user_uid, user_gid):
            def sid():
                setgid(user_gid)
                setuid(user_uid)
                logger.debug('set uid=%d gid=%d' % (user_uid, user_gid))
            return sid

        # /etc/sudoer: fdi ALL:(vvpp) NOPASSWD: ALL
        # gpasswd -a xxxx fdi
        # cmd = ['sudo', '-u', as_user, 'bash', '-l', '-c'] + cmd
        # cmd = ['sudo', '-u', as_user] + cmd

        shell = False
        logger.debug('Popen %s env:%s uid: %d gid:%d' %
                     (str(cmd), lls(env, 40), user_uid, user_gid))
        proc = Popen(cmd, executable=executable,
                     stdin=pty, stdout=pty, stderr=pty,
                     preexec_fn=None,
                     cwd=pwdir,
                     env=env, shell=shell,
                     start_new_session=False,
                     encoding='utf-8')  # , universal_newlines=True)
    except Exception as e:
        msg = trbk(e)
        logger.info('Running popen got exception. '+msg)
        return {'returncode': -1, 'message': msg}

    sta = {'command': str(cmd)}
    try:
        sta['stdout'], sta['stderr'] = proc.communicate(timeout=timeout)
        sta['returncode'] = proc.returncode
        msg = 'Successful.'
    except TimeoutExpired:
        # The child process is not killed if the timeout expires,
        # so in order to cleanup properly a well-behaved application
        # should kill the child process and finish communication
        # https://docs.python.org/3.6/library/subprocess.html?highlight=subprocess#subprocess.Popen.communicate
        proc.kill()
        # kill subprocess, too
        # os.killpg(os.getpgid(proc.pid), signal.SIGTERM)
        sta['stdout'], sta['stderr'] = proc.communicate()
        sta['returncode'] = proc.returncode
        msg = 'PID %d is terminated after pre-set timed-out %d sec.' % (
            proc.pid, timeout)
    sta['message'] = msg
    return sta


def kill_processes(pid):
    """Kills parent and children processess

    https://codeigo.com/python/kill-subprocess
    """
    import psutil
    parent = psutil.Process(pid)
    # kill all the child processes
    for child in parent.children(recursive=True):
        print(child)
        child.kill()
    # kill the parent process
    print(parent)
    parent.kill()



# -*- coding: utf-8 -*-
import logging
# create logger
logger = logging.getLogger(__name__)
#logger.debug('level %d' %  (logger.getEffectiveLevel()))


class Typed():
    """ An object of specifiable kinds.

    This class is for parameters, not for products/datasets to make 'type' MDP.
    """

    def __init__(self, typ_=None, **kwds):
        """ Has a type.
        Parameters
        ----------

        Returns
        -------
        """

        # print(__name__ + str(kwds))
        super().__init__(**kwds)
        self.setType(typ_)

    @property
    def type(self):
        """ for property getter
        Parameters
        ----------

        Returns
        -------
        """
        return self.getType()

    @type.setter
    def type(self, typ_):
        """ Must be in ParameterTypes.
        Parameters
        ----------
        Returns
        -------
        """
        self.setType(typ_)

    def getType(self):
        """ Returns the actual type that is allowed for the value
        of this Parameter.
        Parameters
        ----------

        Returns
        -------
        """
        return self._type

    def setType(self, typ_):
        """ Replaces the current type of this parameter.
        Parameters
        ----------

        Returns
        -------
        """
        self._type = typ_



# -*- coding: utf-8 -*-

import logging
# create logger
logger = logging.getLogger(__name__)
#logger.debug('level %d' %  (logger.getEffectiveLevel()))


class VAttribute():
    """ Attributes are stored in instance variables.

    Variable names start with a '_'.
    """

    def __init__(self, **kwds):
        """
        Parameter
        ---------

        Returns
        -------
        """
        super().__init__(**kwds)

    def __getattribute__(self, name):
        """ Returns the '_name'd attitute value.
        """
        if name.startswith('_'):
            return super().__getattribute__(name)
        else:
            return super().__getattribute__(f'_{name}')

    def __setattr__(self, name, value):
        """ Set the '_name'd attitute value to `value`.
        """
        if name.startswith('_'):
            super().__setattr__(name, value)
        else:
            super().__setattr__(f'_{name}', value)



# -*- coding: utf-8 -*-
out_tree = """tree out_tree
├── meta                                          <MetaData>
│   └── listeners                              <ListenerSet>
├── measurements                          <CompositeDataset>
│   ├── meta                                      <MetaData>
│   │   └── listeners                          <ListenerSet>
│   ├── Time_Energy_Pos               <TableDataset> [5, 20]
│   │   ├── meta                                  <MetaData>
│   │   │   └── listeners                      <ListenerSet>
│   │   ├── Time                               <Column> [20]
│   │   ├── Energy                             <Column> [20]
│   │   ├── Error                              <Column> [20]
│   │   ├── y                                  <Column> [20]
│   │   └── z                                  <Column> [20]
│   ├── calibration                  <ArrayDataset> [11, 11]
│   └── dset                                           <str>
├── Environment Temperature               <ArrayDataset> [7]
├── Browse                                <image/png> [5976]
├── refs                                      <RefContainer>
│   ├── a_reference                             <ProductRef>
│   │   └── urnobj                                     <Urn>
│   │       └── urn                                    <str>
│   └── a_different_name                        <ProductRef>
├── history                            <TableDataset> [2, 0]
│   ├── meta                                      <MetaData>
│   │   └── listeners                          <ListenerSet>
│   ├── name                                    <Column> [0]
│   └── reference                               <Column> [0]
└── listeners                                  <ListenerSet>
├── meta                                          <MetaData>
│   ├── description                                 <string>
│   │   ├── description                                <str>
│   │   ├── default                                    <str>
│   │   ├── value                                      <str>
│   │   ├── valid                                 <NoneType>
│   │   └── typecode                                   <str>
│   ├── type                                        <string>
│   │   ├── description                                <str>
│   │   ├── default                                    <str>
│   │   ├── value                                      <str>
│   │   ├── valid                                 <NoneType>
│   │   └── typecode                                   <str>
│   ├── level                                       <string>
│   │   ├── description                                <str>
│   │   ├── default                                    <str>
│   │   ├── value                                      <str>
│   │   ├── valid                                 <NoneType>
│   │   └── typecode                                   <str>
│   ├── creator                                     <string>
│   │   ├── description                                <str>
│   │   ├── default                                    <str>
│   │   ├── value                                      <str>
│   │   ├── valid                                 <NoneType>
│   │   └── typecode                                   <str>
│   ├── creationDate                              <finetime>
│   │   ├── description                                <str>
│   │   ├── default                               <FineTime>
│   │   │   ├── tai                                    <int>
│   │   │   └── format                                 <str>
│   │   ├── value                                 <FineTime>
│   │   │   ├── tai                                    <int>
│   │   │   └── format                                 <str>
│   │   ├── valid                                 <NoneType>
│   │   └── typecode                                   <str>
│   ├── rootCause                                   <string>
│   │   ├── description                                <str>
│   │   ├── default                                    <str>
│   │   ├── value                                      <str>
│   │   ├── valid                                 <NoneType>
│   │   └── typecode                                   <str>
│   ├── version                                     <string>
│   │   ├── description                                <str>
│   │   ├── default                                    <str>
│   │   ├── value                                      <str>
│   │   ├── valid                                 <NoneType>
│   │   └── typecode                                   <str>
│   ├── FORMATV                                     <string>
│   │   ├── description                                <str>
│   │   ├── default                                    <str>
│   │   ├── value                                      <str>
│   │   ├── valid                                 <NoneType>
│   │   └── typecode                                   <str>
│   ├── speed                                       <vector>
│   │   ├── description                                <str>
│   │   ├── type                                       <str>
│   │   ├── default                               <NoneType>
│   │   ├── value                                   <Vector>
│   │   │   ├── components                            <list>
│   │   │   ├── unit                              <NoneType>
│   │   │   └── typecode                          <NoneType>
│   │   ├── valid                                     <list>
│   │   │   ├── 0                                     <list>
│   │   │   │   └── 0                                 <list>
│   │   │   └── 1                                     <list>
│   │   │       └── 0                                 <list>
│   │   ├── unit                                       <str>
│   │   └── typecode                              <NoneType>
│   └── listeners                              <ListenerSet>
├── measurements                          <CompositeDataset>
│   ├── meta                                      <MetaData>
│   │   └── listeners                          <ListenerSet>
│   ├── Time_Energy_Pos               <TableDataset> [5, 20]
│   │   ├── meta                                  <MetaData>
│   │   │   ├── description                         <string>
│   │   │   │   ├── description                        <str>
│   │   │   │   ├── default                            <str>
│   │   │   │   ├── value                              <str>
│   │   │   │   ├── valid                         <NoneType>
│   │   │   │   └── typecode                           <str>
│   │   │   ├── shape                                 <list>
│   │   │   │   ├── description                        <str>
│   │   │   │   ├── type                               <str>
│   │   │   │   ├── default                           <list>
│   │   │   │   ├── value                             <list>
│   │   │   │   ├── valid                         <NoneType>
│   │   │   │   ├── unit                               <str>
│   │   │   │   └── typecode                           <str>
│   │   │   ├── type                                <string>
│   │   │   │   ├── description                        <str>
│   │   │   │   ├── default                            <str>
│   │   │   │   ├── value                              <str>
│   │   │   │   ├── valid                         <NoneType>
│   │   │   │   └── typecode                           <str>
│   │   │   ├── version                             <string>
│   │   │   │   ├── description                        <str>
│   │   │   │   ├── default                            <str>
│   │   │   │   ├── value                              <str>
│   │   │   │   ├── valid                         <NoneType>
│   │   │   │   └── typecode                           <str>
│   │   │   ├── FORMATV                             <string>
│   │   │   │   ├── description                        <str>
│   │   │   │   ├── default                            <str>
│   │   │   │   ├── value                              <str>
│   │   │   │   ├── valid                         <NoneType>
│   │   │   │   └── typecode                           <str>
│   │   │   └── listeners                      <ListenerSet>
│   │   ├── Time                               <Column> [20]
│   │   ├── Energy                             <Column> [20]
│   │   ├── Error                              <Column> [20]
│   │   ├── y                                  <Column> [20]
│   │   └── z                                  <Column> [20]
│   ├── calibration                  <ArrayDataset> [11, 11]
│   └── dset                                           <str>
├── Environment Temperature               <ArrayDataset> [7]
├── Browse                                <image/png> [5976]
├── refs                                      <RefContainer>
│   ├── a_reference                             <ProductRef>
│   │   └── urnobj                                     <Urn>
│   │       └── urn                                    <str>
│   └── a_different_name                        <ProductRef>
├── history                            <TableDataset> [2, 0]
│   ├── meta                                      <MetaData>
│   │   ├── description                             <string>
│   │   │   ├── description                            <str>
│   │   │   ├── default                                <str>
│   │   │   ├── value                                  <str>
│   │   │   ├── valid                             <NoneType>
│   │   │   └── typecode                               <str>
│   │   ├── shape                                     <list>
│   │   │   ├── description                            <str>
│   │   │   ├── type                                   <str>
│   │   │   ├── default                               <list>
│   │   │   ├── value                                 <list>
│   │   │   ├── valid                             <NoneType>
│   │   │   ├── unit                                   <str>
│   │   │   └── typecode                               <str>
│   │   ├── type                                    <string>
│   │   │   ├── description                            <str>
│   │   │   ├── default                                <str>
│   │   │   ├── value                                  <str>
│   │   │   ├── valid                             <NoneType>
│   │   │   └── typecode                               <str>
│   │   ├── version                                 <string>
│   │   │   ├── description                            <str>
│   │   │   ├── default                                <str>
│   │   │   ├── value                                  <str>
│   │   │   ├── valid                             <NoneType>
│   │   │   └── typecode                               <str>
│   │   ├── FORMATV                                 <string>
│   │   │   ├── description                            <str>
│   │   │   ├── default                                <str>
│   │   │   ├── value                                  <str>
│   │   │   ├── valid                             <NoneType>
│   │   │   └── typecode                               <str>
│   │   └── listeners                          <ListenerSet>
│   ├── name                                    <Column> [0]
│   └── reference                               <Column> [0]
└── listeners                                  <ListenerSet>
|__ meta                                          <MetaData>
|   |__ description                                 <string>
|   |   |__ description                                <str>
|   |   |__ default                                    <str>
|   |   |__ value                                      <str>
|   |   |__ valid                                 <NoneType>
|   |   \__ typecode                                   <str>
|   |__ type                                        <string>
|   |   |__ description                                <str>
|   |   |__ default                                    <str>
|   |   |__ value                                      <str>
|   |   |__ valid                                 <NoneType>
|   |   \__ typecode                                   <str>
|   |__ level                                       <string>
|   |   |__ description                                <str>
|   |   |__ default                                    <str>
|   |   |__ value                                      <str>
|   |   |__ valid                                 <NoneType>
|   |   \__ typecode                                   <str>
|   |__ creator                                     <string>
|   |   |__ description                                <str>
|   |   |__ default                                    <str>
|   |   |__ value                                      <str>
|   |   |__ valid                                 <NoneType>
|   |   \__ typecode                                   <str>
|   |__ creationDate                              <finetime>
|   |   |__ description                                <str>
|   |   |__ default                               <FineTime>
|   |   |   |__ tai                                    <int>
|   |   |   \__ format                                 <str>
|   |   |__ value                                 <FineTime>
|   |   |   |__ tai                                    <int>
|   |   |   \__ format                                 <str>
|   |   |__ valid                                 <NoneType>
|   |   \__ typecode                                   <str>
|   |__ rootCause                                   <string>
|   |   |__ description                                <str>
|   |   |__ default                                    <str>
|   |   |__ value                                      <str>
|   |   |__ valid                                 <NoneType>
|   |   \__ typecode                                   <str>
|   |__ version                                     <string>
|   |   |__ description                                <str>
|   |   |__ default                                    <str>
|   |   |__ value                                      <str>
|   |   |__ valid                                 <NoneType>
|   |   \__ typecode                                   <str>
|   |__ FORMATV                                     <string>
|   |   |__ description                                <str>
|   |   |__ default                                    <str>
|   |   |__ value                                      <str>
|   |   |__ valid                                 <NoneType>
|   |   \__ typecode                                   <str>
|   |__ speed                                       <vector>
|   |   |__ description                                <str>
|   |   |__ type                                       <str>
|   |   |__ default                               <NoneType>
|   |   |__ value                                   <Vector>
|   |   |   |__ components                            <list>
|   |   |   |__ unit                              <NoneType>
|   |   |   \__ typecode                          <NoneType>
|   |   |__ valid                                     <list>
|   |   |   |__ 0                                     <list>
|   |   |   |   \__ 0                                 <list>
|   |   |   \__ 1                                     <list>
|   |   |       \__ 0                                 <list>
|   |   |__ unit                                       <str>
|   |   \__ typecode                              <NoneType>
|   \__ listeners                              <ListenerSet>
|__ measurements                          <CompositeDataset>
|   |__ meta                                      <MetaData>
|   |   \__ listeners                          <ListenerSet>
|   |__ Time_Energy_Pos               <TableDataset> [5, 20]
|   |   |__ meta                                  <MetaData>
|   |   |   |__ description                         <string>
|   |   |   |   |__ description                        <str>
|   |   |   |   |__ default                            <str>
|   |   |   |   |__ value                              <str>
|   |   |   |   |__ valid                         <NoneType>
|   |   |   |   \__ typecode                           <str>
|   |   |   |__ shape                                 <list>
|   |   |   |   |__ description                        <str>
|   |   |   |   |__ type                               <str>
|   |   |   |   |__ default                           <list>
|   |   |   |   |__ value                             <list>
|   |   |   |   |__ valid                         <NoneType>
|   |   |   |   |__ unit                               <str>
|   |   |   |   \__ typecode                           <str>
|   |   |   |__ type                                <string>
|   |   |   |   |__ description                        <str>
|   |   |   |   |__ default                            <str>
|   |   |   |   |__ value                              <str>
|   |   |   |   |__ valid                         <NoneType>
|   |   |   |   \__ typecode                           <str>
|   |   |   |__ version                             <string>
|   |   |   |   |__ description                        <str>
|   |   |   |   |__ default                            <str>
|   |   |   |   |__ value                              <str>
|   |   |   |   |__ valid                         <NoneType>
|   |   |   |   \__ typecode                           <str>
|   |   |   |__ FORMATV                             <string>
|   |   |   |   |__ description                        <str>
|   |   |   |   |__ default                            <str>
|   |   |   |   |__ value                              <str>
|   |   |   |   |__ valid                         <NoneType>
|   |   |   |   \__ typecode                           <str>
|   |   |   \__ listeners                      <ListenerSet>
|   |   |__ Time                               <Column> [20]
|   |   |__ Energy                             <Column> [20]
|   |   |__ Error                              <Column> [20]
|   |   |__ y                                  <Column> [20]
|   |   \__ z                                  <Column> [20]
|   |__ calibration                  <ArrayDataset> [11, 11]
|   \__ dset                                           <str>
|__ Environment Temperature               <ArrayDataset> [7]
|__ Browse                                <image/png> [5976]
|__ refs                                      <RefContainer>
|   |__ a_reference                             <ProductRef>
|   |   \__ urnobj                                     <Urn>
|   |       \__ urn                                    <str>
|   \__ a_different_name                        <ProductRef>
|__ history                            <TableDataset> [2, 0]
|   |__ meta                                      <MetaData>
|   |   |__ description                             <string>
|   |   |   |__ description                            <str>
|   |   |   |__ default                                <str>
|   |   |   |__ value                                  <str>
|   |   |   |__ valid                             <NoneType>
|   |   |   \__ typecode                               <str>
|   |   |__ shape                                     <list>
|   |   |   |__ description                            <str>
|   |   |   |__ type                                   <str>
|   |   |   |__ default                               <list>
|   |   |   |__ value                                 <list>
|   |   |   |__ valid                             <NoneType>
|   |   |   |__ unit                                   <str>
|   |   |   \__ typecode                               <str>
|   |   |__ type                                    <string>
|   |   |   |__ description                            <str>
|   |   |   |__ default                                <str>
|   |   |   |__ value                                  <str>
|   |   |   |__ valid                             <NoneType>
|   |   |   \__ typecode                               <str>
|   |   |__ version                                 <string>
|   |   |   |__ description                            <str>
|   |   |   |__ default                                <str>
|   |   |   |__ value                                  <str>
|   |   |   |__ valid                             <NoneType>
|   |   |   \__ typecode                               <str>
|   |   |__ FORMATV                                 <string>
|   |   |   |__ description                            <str>
|   |   |   |__ default                                <str>
|   |   |   |__ value                                  <str>
|   |   |   |__ valid                             <NoneType>
|   |   |   \__ typecode                               <str>
|   |   \__ listeners                          <ListenerSet>
|   |__ name                                    <Column> [0]
|   \__ reference                               <Column> [0]
\__ listeners                                  <ListenerSet>"""



# -*- coding: utf-8 -*-
from copy import deepcopy
import logging

logger = logging.getLogger(__name__)


class Copyable():
    """ Interface for objects that can make a copy of themselves. """

    def __init__(self, **kwds):
        """

        Parameters
        ----------

        Returns
        -------
        """

        super().__init__(**kwds)

    def copy(self):
        """ Makes a deep copy of itself.
        Parameters
        ----------

        Returns
        -------
        """
        return deepcopy(self)



# -*- coding: utf-8 -*-

from .metadata import Parameter
from .typecoded import Typecoded

from copy import copy
from collections import OrderedDict
import logging
# create logger
logger = logging.getLogger(__name__)
#logger.debug('level %d' %  (logger.getEffectiveLevel()))


class StringParameter(Parameter, Typecoded):
    """ has a unicode string as the value, a typecode for length and char.
    """

    def __init__(self,
                 value=None,
                 description='UNKNOWN',
                 default=None,
                 valid=None,
                 typecode='B',
                 **kwds):

        typ_ = kwds.pop('typ_', 'string')
        # collect args-turned-local-variables.
        args = copy(locals())
        args.pop('__class__', None)
        args.pop('kwds', None)
        args.pop('self', None)
        args.update(kwds)

        self.setTypecode(typecode)
        super().__init__(
            value=value, description=description, typ_=typ_, default=default, valid=valid, typecode=typecode)
        # Must overwrite the self._all_attrs set by supera()
        self._all_attrs = args

    def __getstate__(self):
        """ Can be encoded with serializableEncoder """
        return OrderedDict(
            description=self.description if hasattr(
                self, 'description') else '',
            default=self._default,
            value=self._value if hasattr(self, '_value') else None,
            valid=self._valid,
            typecode=self._typecode)



# -*- coding: utf-8 -*-

# from ..route.pools import pools_api
from ...utils.common import (logging_ERROR,
                             logging_WARNING,
                             logging_INFO,
                             logging_DEBUG
                             )

from flask import (abort,
                   Blueprint,
                   flash,
                   g,
                   make_response,
                   render_template,
                   request,
                   session)
from werkzeug.security import generate_password_hash, check_password_hash
from flask import current_app
from flask_httpauth import HTTPBasicAuth

import datetime
import time
import functools
import logging

logger = logging.getLogger(__name__)

auth = HTTPBasicAuth()

SESSION = True
""" Enable session. """


LOGIN_TMPLT = ''  # 'user/login.html'
""" Set LOGIN_TMPLT to '' to disable the login page."""


user = Blueprint('user', __name__)


class User():

    def __init__(self, username,
                 password=None,
                 hashed_password=None,
                 roles=['read_only']):

        global logger

        self.username = username
        if hashed_password:
            if password:
                if logger.isEnabledFor(logging_WARNING):
                    logger.warning(
                        'Both password and hashed_password are given for %s. Password is igored.' % username)
                password = None
        elif password:
            hashed_password = self.hash_of(password)
        else:
            raise ValueError(
                'No password and no hashed_password given for ' + username)
        self.password = password
        self.registered_on = datetime.datetime.now()

        self.hashed_password = hashed_password
        self.roles = (roles,) if issubclass(
            roles.__class__, str) else tuple(roles)
        self.authenticated = False

    @functools.lru_cache(maxsize=1024)
    def is_correct_password(self, plaintext_password):

        return check_password_hash(self.hashed_password, plaintext_password)

    @staticmethod
    @functools.lru_cache(maxsize=64)
    def hash_of(s):
        return generate_password_hash(s)

    def __repr__(self):
        return f'<User: {self.username}>'

    def getCacheInfo(self):
        info = {}
        for i in ['is_correct_password', 'hash_of']:
            info[i] = getattr(self, i).cache_info()
        return info


def get_names2roles_mapping(pc):
    # pc is pnsconfig from config filea
    mapping = {'read_write': [], 'read_only': []}
    for authtype in ('rw_user', 'ro_user'):
        unames = pc[authtype]
        unames = unames if isinstance(unames, list) else [unames]
        for n in unames:
            if authtype == 'rw_user':
                mapping['read_write'].append(n)
            else:
                mapping['read_only'].append(n)
    return mapping


NAMES2ROLES = None


def getUsers(pc):
    """ Returns the USER DB from `config.py` ro local config file.

    Allows multiple user under the same role"""

    global NAMES2ROLES
    if NAMES2ROLES is None:
        NAMES2ROLES = get_names2roles_mapping(pc)
    users = dict(((u, User(u, None, hp,
                           [r for r, n in NAMES2ROLES.items() if u in n]))
                  for u, hp in ((pc['rw_user'], pc['rw_pass']),
                                (pc['ro_user'], pc['ro_pass']))
                  ))
    return users


SES_DBG = 0
""" debug msg for session """

if SESSION:
    @user.before_app_request
    def load_logged_in_user():
        logger = current_app.logger
        user_id = session.get('user_id')
        if SES_DBG and logger.isEnabledFor(logging_DEBUG):
            headers = dict(request.headers)
            cook = dict(request.cookies)

            logger.debug('S:%x "%s", %s, %s' %
                         (id(session), str(user_id),
                          str(headers.get('Authorization', '')),
                          str(cook.get('session', '')[:6])))
        if user_id is None:
            g.user = None
        else:
            g.user = current_app.config['USERS'][user_id]


@auth.get_user_roles
def get_user_roles(user):
    if issubclass(user.__class__, User):
        return user.roles
    else:
        return None


######################################
####  /login GET, POST  ####
######################################


@ user.route('/login/', methods=['GET', 'POST'])
@ user.route('/login', methods=['GET', 'POST'])
# @ auth.login_required(role=['read_only', 'read_write'])
def login():
    """ Logging in on the server.

    :return: response made from http code, poolurl, message
    """
    global logger
    logger = current_app.logger
    ts = time.time()
    acu = auth.current_user()

    try:
        reqanm = request.authorization['username']
        reqanm = request.authorization['passwd']
    except (AttributeError, TypeError):
        reqanm = reqaps = ''
    if logger.isEnabledFor(logging_DEBUG):
        msg = 'LOGIN meth=%s req_auth_nm= "%s"' % (request.method, reqanm)
        logger.debug(msg)

    if request.method == 'POST':
        rnm = request.form.get('username', None)
        rpas = request.form.get('password', None)
        if logger.isEnabledFor(logging_DEBUG):
            logger.debug(f'Request form {rnm}')

        if not (rpas and rnm):
            msg = 'Bad username or password posted %s' % str(rnm)
            if logger.isEnabledFor(logging_WARNING):
                logger.warning(msg)
            if reqanm and reqaps:
                if logger.isEnabledFor(logging_WARNING):
                    msg = f'Username {reqanm} and pswd in auth header used.'
                    logger.warning(msg)
                rnm, rpas = reqanm, reqaps

        vp = verify_password(rnm, rpas, check_session=False)
        if vp in (False, None):
            if logger.isEnabledFor(logging_DEBUG):
                msg = f'Verifying {rnm} with password failed.'
                logger.debug(msg)
        else:
            if SESSION:
                session.clear()
                session['user_id'] = rnm
                session.new = True
                session.modified = True
            msg = 'User %s logged-in %s.' % (rnm, vp.roles)
            if logger.isEnabledFor(logging_DEBUG):
                logger.debug(msg)
            # return redirect(url_for('pools.get_pools_url'))
            if SESSION:
                flash(msg)
            from ..route.httppool_server import resp
            return resp(200, 'OK.', msg, ts, req_auth=True)
    elif request.method == 'GET':
        if logger.isEnabledFor(logging_DEBUG):
            logger.debug('start login')
    else:
        if logger.isEnabledFor(logging_ERROR):
            logger.error('How come the method is ' + request.method)
    if LOGIN_TMPLT:
        if SESSION:
            flash(msg)
        else:
            if logger.isEnabledFor(logging_INFO):
                logger.info(msg)
        return make_response(render_template(LOGIN_TMPLT))
    else:
        abort(401)

######################################
####  /user/logout GET, POST  ####
######################################


@ user.route('/logout/', methods=['GET', 'POST'])
@ user.route('/logout', methods=['GET', 'POST'])
# @ auth.login_required(role=['read_only', 'read_write'])
def logout():
    """ Logging in on the server.

    :return: response made from http code, poolurl, message
    """

    logger = current_app.logger
    ts = time.time()
    if logger.isEnabledFor(logging_DEBUG):
        logger.debug('logout')
    # session.get('user_id') is the name

    if SESSION and hasattr(g, 'user') and hasattr(g.user, 'username'):
        nm, rl = g.user.username, g.user.roles
        msg = 'User %s logged-out %s.' % (nm, rl)
        res = 'OK. Bye, %s (%s).' % (nm, rl)
    else:
        msg = 'User logged-out.'
        res = 'OK. Bye.'
    if logger.isEnabledFor(logging_DEBUG):
        logger.debug(msg)
    if SESSION:
        session.clear()
        g.user = None
        session.new = True
        session.modified = True

    from ..route.httppool_server import resp

    return resp(200, res, msg, ts)


@auth.verify_password
def verify_password(username, password, check_session=True):
    """ Call back.

    ref: https://flask-httpauth.readthedocs.io/en/latest/index.html

        must return the user object if credentials are valid,
        or True if a user object is not available. In case of
        failed authentication, it should return None or False.

    `check_session`=`True` ('u/k' means unknown)

    =========== ============= ======= ========== ========= ==================
    state          `session`   `g`     username  password      action
    =========== ============= ======= ========== ========= ==================
    no Session  no 'user_id'          not empty  valid     new session, r/t new u
    no Session  no 'user_id'          not empty  invalid   login, r/t `False`
    no Session  no 'user_id'          ''                   r/t None
    no Session  no 'user_id'          None, u/k            login, r/t `False`
    no SESSION  not enabled           not empty  cleartext approve
   In session  w/ 'user_id'  ''|None not empty  valid      new session, r/t new u
    ..                                not same
    In session  w/ 'user_id'          not empty  invalid   login, return `False`
    In session  w/ 'user_id'  user    None ""              login, return `False`
    ..                                u/k
    =========== ============= ======= ========== ========= ==================

    `check_session`=`False`

    ========== ========= =========  ================
     in USERS  username  password    action
    ========== ========= =========  ================
     False                          return False
     True      not empty  valid     return user
     True      not empty  invalid   return False
               ''                   return None
               None                 return False
    ========== ========= =========  ================

    No SESSION:

    > return `True`
    """
    logger = current_app.logger

    if SES_DBG and logger.isEnabledFor(logging_DEBUG):
        logger.debug('%s %s %s %s' % (username, len(password) * '*',
                     'chk' if check_session else 'nochk',
                                      'Se' if SESSION else 'noSe'))

    if check_session:
        if SESSION:
            has_session = 'user_id' in session and hasattr(
                g, 'user') and g.user is not None
            if has_session:
                user = g.user
                if SES_DBG and logger.isEnabledFor(logging_DEBUG):
                    logger.debug(f'g.usr={user.username}')
                gname = user.username
                newu = current_app.config['USERS'].get(username, None)
                # first check if the username is actually unchanged and valid
                if newu is not None and newu.is_correct_password(password):
                    if gname == username:
                        if logger.isEnabledFor(logging_DEBUG):
                            logger.debug(f"Same session.")
                        return user
                        #################
                    else:
                        if logger.isEnabledFor(logging_INFO):
                            logger.info(f"New session {username}.")
                        session.clear()
                        session['user_id'] = username
                        session.modified = True
                        return newu
                        #################
                if logger.isEnabledFor(logging_DEBUG):
                    logger.debug(
                        f"Unknown {username} or Null or anonymous user, or new user '{username}' has invalid password.")
                return False
                #################
            else:
                # SESSION enabled but has not valid user_id
                if logger.isEnabledFor(logging_DEBUG):
                    logger.debug('no session.'
                                 'has %s "user_id". %s g. g.user= %s' % (
                                     ('' if 'user_id' in session else 'no'),
                                     ('' if hasattr(g, 'user') else 'no'), (g.get('user', 'None'))))
                if username == '':
                    if logger.isEnabledFor(logging_DEBUG):
                        logger.debug(f"Anonymous user.")
                    return None
                    #################
                newu = current_app.config['USERS'].get(username, None)
                if newu is None:

                    if logger.isEnabledFor(logging_DEBUG):
                        logger.debug(f"Unknown user {username}")
                    return False
                    #################
                if newu.is_correct_password(password):
                    if logger.isEnabledFor(logging_INFO):
                        logger.info(
                            f'Approved new user {username}. Start new session')
                    session.clear()
                    session['user_id'] = username
                    session.modified = True
                    return newu
                    #################
                else:
                    if logger.isEnabledFor(logging_DEBUG):
                        logger.debug(
                            f"new user '{username}' has invalid password.")
                    return False
                    #################
        else:
            # SESSION not enabled. Use clear text passwd
            newu = current_app.config['USERS'].get(username, None)
            if newu and newu.is_correct_password(password):
                if logger.isEnabledFor(logging_INFO):
                    logger.info('Approved new user {username} w/o session')
                return newu
                #################
            else:
                if logger.isEnabledFor(logging_DEBUG):
                    logger.debug(
                        f"Null or anonymous user, or new user '{username}' has invalid password.")
                return False
                #################
    else:
        # check_session is False. called by login to check formed name/pass
        if username == '':
            if logger.isEnabledFor(logging_DEBUG):
                logger.debug('LOGIN: check anon')
            return None
            #################
        newu = current_app.config['USERS'].get(username, None)
        if newu is None:
            if logger.isEnabledFor(logging_DEBUG):
                logger.debug(f"LOGIN: Unknown user {username}")
            return False
            #################
        if newu.is_correct_password(password):
            if logger.isEnabledFor(logging_INFO):
                logger.info('LOGIN Approved {username}')
            return newu
            #################
        else:
            if logger.isEnabledFor(logging_DEBUG):
                logger.debug('LOGIN False for {username}')
            return False
            #################


######################################
####  /register GET, POST  ####
######################################


@user.route('/register', methods=('GET', 'POST'))
def register():
    ts = time.time()
    from ..route.httppool_server import resp
    return resp(300, 'FAILED', 'Not available.', ts)


if LOGIN_TMPLT:
    @auth.error_handler
    def handle_auth_error_codes(error=401):
        """ if verify_password returns False, this gets to run.

        Note that this is decorated with flask_httpauth's `error_handler`, not flask's `errorhandler`.
        """
        if error in [401, 403]:
            # send a login page
            current_app.logger.debug("Error %d. Start login page..." % error)
            page = make_response(render_template(LOGIN_TMPLT))
            return page
        else:
            raise ValueError('Must be 401 or 403. Nor %s' % str(error))


# open text passwd
# @auth.verify_password
# def verify(username, password):
#     """This function is called to check if a username /
#     password combination is valid.
#     """
#     pc = current_app.config['PC']
#     if not (username and password):
#         return False
#     return username == pc['username'] and password == pc['password']

    # if 0:
    #        pass
    # elif username == pc['auth_user'] and password == pc['auth_pass']:

    # else:
    #     password = str2md5(password)
    #     try:
    #         conn = mysql.connector.connect(host = pc['mysql']['host'], port=pc['mysql']['port'], user =pc['mysql']['user'], password = pc['mysql']['password'], database = pc['mysql']['database'])
    #         if conn.is_connected():
    #             current_app.logger.info("connect to db successfully")
    #             cursor = conn.cursor()
    #             cursor.execute("SELECT * FROM userinfo WHERE userName = '" + username + "' AND password = '" + password + "';" )
    #             record = cursor.fetchall()
    #             if len(record) != 1:
    #                 current_app.logger.info("User : " + username + " auth failed")
    #                 conn.close()
    #                 return False
    #             else:
    #                 conn.close()
    #                 return True
    #         else:
    #             return False
    #     except Error as e:
    #         current_app.logger.error("Connect to database failed: " +str(e))


def login_required1(view):
    @functools.wraps(view)
    def wrapped_view(**kwargs):
        if g.user is None:
            return 401, 'FAILED', "This operation needs authorization."

        return view(**kwargs)

    return wrapped_view



# -*- coding: utf-8 -*-

from ..dataset.arraydataset import ArrayDataset, Column
from ..dataset.tabledataset import TableDataset
from ..dataset.datatypes import Vector
from ..dataset.finetime import FineTime
from ..dataset.metadata import Parameter
from ..dataset.attributable import Reserved_Property_Names
from .common import bstr, lls

from itertools import chain, islice, repeat
import logging
from collections import OrderedDict
from collections.abc import Mapping, Sequence
from pprint import pprint
import array
import time
import io
import statistics

# create logger
logger = logging.getLogger(__name__)
# logger.debug('level %d' %  (logger.getEffectiveLevel()))


"""
Aaron Hall
https://stackoverflow.com/a/59109706
"""

SIMPLE_TYPES = (int, float, complex, str, bytes, bool, type(None),
                FineTime, Vector)
# prefix components:
PREFIXES = {'line': (
    # space
    '    ',
    # branch
    '│   ',
    # pointers
    # tee =
    '├── ',
    # last
    '└── '
),
    'ascii': (
    # space
    '    ',
    # branch
    '|   ',
    # pointers
    # tee =
    '|__ ',
    # last
    '\__ '
), }


def tree(data_object, level=0, style='line', prefix='', seen=None):
    """A recursive generator, given an object object
    will yield a visual tree structure line by line
    with each line prefixed by the same characters.

    :style: can be ```line``` pr ```ascii```.
    """

    space, branch, tee, last = PREFIXES[style]
    if seen is None:
        seen = []
    elif id(data_object) in seen:
        return
    seen.append(id(data_object))
    # data is a Sequence
    if issubclass(data_object.__class__, Sequence):
        contents = list((n, v) for n, v in enumerate(data_object)
                        if not issubclass(v.__class__, SIMPLE_TYPES))
        if len(contents) == 0:
            return
    elif hasattr(data_object, '__getstate__'):
        # state variables. Change names "_ATTR_xxx" to "xxx". Filter out Parameter
        contents = list(((n[6:] if n.startswith('_ATTR_') else n), v)
                        for n, v in data_object.__getstate__().items() if
                        level > 0 or not issubclass(v.__class__, Parameter))
    elif issubclass(data_object.__class__, Mapping):
        # data is MutableMapping of simple values
        contents = list((n, v) for n, v in
                        data_object.items() if
                        not (n.startswith('_') or n in Reserved_Property_Names
                             ))
    elif hasattr(data_object, '__dict__'):
        # or properties
        contents = list((n, v) for n, v in
                        vars(data_object).items() if
                        not (n.startswith('_') or n in Reserved_Property_Names
                             ))
    # contents each get pointers that are ├── with a final └── :
    pointers = [tee] * (len(contents) - 1) + [last]
    for pointer, name_value in zip(pointers, contents):
        v = name_value[1]
        vc = v.__class__
        shp = (' %s' % str(v.shape)) if hasattr(v, 'shape') else ''
        typ = '%s' % (str(v.type) if hasattr(v, 'type') else vc.__name__)
        # format output line
        ts = '<%s>%s' % (typ, shp)
        line = prefix + pointer + str(name_value[0])
        yield '%s%s%s' % (line, ' ' * max(1, 60-len(line)-len(ts)), ts)

        if issubclass(vc, (str, bytes)) or \
           issubclass(vc, (ArrayDataset)) and hasattr(v, 'typecode'):
            pass
        elif issubclass(vc, (Mapping, Sequence)) or hasattr(v, '__dict__'):
            # extend the prefix and recurse:
            extension = branch if pointer == tee else space
            # i.e. space because last, └── , above so no more |
            yield from tree(v, level=level, style=style, prefix=prefix+extension, seen=seen)



import fdi.dataset.namespace
from fdi.dataset.namespace import NameSpace_meta, Lazy_Loading_ChainMap, Load_Failed

from collections import ChainMap
import pytest
import importlib
import copy
import sys
import logging

# from test_dataset import check_Product

if sys.version_info[0] >= 3:  # + 0.1 * sys.version_info[1] >= 3.3:
    PY3 = True
else:
    PY3 = False

# create logger
logger = logging.getLogger(__name__)
# logger.debug('level %d' %  (logger.getEffectiveLevel()))


@pytest.fixture(scope='function')
def NSmeta():
    yield fdi.dataset.namespace.NameSpace_meta
    importlib.reload(fdi.dataset.namespace)


@pytest.fixture(scope='function')
def claz():
    import fdi.dataset.classes
    importlib.reload(fdi.dataset.classes)
    from fdi.dataset.classes import Classes
    return Classes


def test_Lazy(NSmeta):
    d = object()
    a = {"a": d, "f": 8}
    b = {"c": 55}
    e = {"ee": 0, "c": 22}
    v = Lazy_Loading_ChainMap(a, b, extensions=[e])
    assert v
    lab = len(a)+len(b)
    assert len(v.initial) == lab
    assert len(v.cache) == 0
    # first access removes 'a' from initial
    assert v["a"] is d
    assert len(a) == 2
    assert len(b) == 1
    assert len(v.initial) == lab-1
    # e has only 1 unique key. len =  + (len(e)-1)
    assert len(v) == len(v.initial) + len(v.cache) + (len(e)-1)
    assert len(v.cache) == 1
    assert v.cache['a'] is d
    # accessing again won't move anything
    assert v["a"] is d
    assert len(a) == 2
    assert len(b) == 1
    assert len(v.initial) == lab-1
    # same with the other map
    v["c"] = 99
    assert len(a) == 2
    # 'c' is still in b
    assert len(b) == 1
    assert 'c' in b
    assert len(v.cache) == 2
    assert 'c' in v.cache
    # 'c' is not in initial
    assert 'c' not in v.initial
    assert len(v.initial) == lab-2
    assert v["c"] == 99
    # access extensions
    assert v["ee"] == 0
    # cache has priority
    assert v["c"] == 99
    # len(v) is correct, the sum of lens of initial and cache
    assert len(v) == lab + (len(e)-1)
    assert len(v) == len(v.initial) + len(v.cache) + (len(e)-1)
    assert v.cache['c'] == 99
    # intentionally feed cache a failure mark
    v.initial['f'] = Load_Failed
    # make sure the above did store 'f' in cache
    assert len(v.cache) == 2
    assert v.initial['f'] is Load_Failed
    assert len(v.cache) == 2
    v.ignore_error = True
    # reading result is None
    assert v['f'] is None
    v.ignore_error = False
    assert len(v.initial) == 1
    assert 'f' in v.initial
    assert 'f' not in v.cache
    assert len(v) == lab + (len(e)-1)
    assert len(v) == len(v.initial) + len(v.cache) + (len(e)-1)
    # exclude
    v.exclude.append('a')

    assert v['a'] is None
    with pytest.raises(KeyError):
        assert v['nonexist'] is None
    v.ignore_error = True
    assert v['nonexist'] is None


def test_NameSpace_func(NSmeta):
    d = object()
    a = {"a": d, "f": 8}
    b = {"c": 55}
    e = {"ee": 0, "c": 22}
    cnt = 0
    from functools import wraps

    def counter(f):
        @wraps
        def w(*a, **k):
            nonlocal cnt
            f(*a, **k)
            cnt += 1
        return w

    def loada(key, mapping, remove=True, exclude=None, ignore_error=False):
        logger.debug(key)
        nonlocal cnt
        cnt += 1
        logger.debug('cnt = %d' % cnt)
        return fdi.dataset.namespace.refloader(key, mapping, remove, exclude=exclude, ignore_error=ignore_error)

    class ns(metaclass=NSmeta,
             sources=[a, b],
             extensions=[e],
             load=loada,
             loadcount=0
             ):
        def __init__(cls, *args, **kwds):
            super().__init__(*args, **kwds)

#
#    class ns(metaclass=nsm):
#        pass
#
    tmap = ns.mapping
    assert cnt == 0

    assert tmap['a'] is d
    assert cnt == 1
    assert tmap.cache['a'] is d
    # e has only 1 unique key. len =  + (len(e)-1)
    assert len(tmap) == len(a)+len(b) + (len(e)-1)
    ns.update({'c': 77})
    # update does not load anything
    assert cnt == 1
    assert list(tmap) == ['ee', 'c', 'f', 'a']
    # getting index does not load anything
    assert cnt == 1
    assert tmap['c'] == 77
    assert len(tmap.initial) == 1
    assert len(tmap) == 3 + (len(e)-1)
    assert cnt == 1
    t = copy.copy(tmap)
    t = copy.deepcopy(tmap)
    assert cnt == 1
    d = dict(tmap)
    assert cnt == 3
    # repr calls each of the maps directly. no loading.
    print(tmap)
    assert cnt == 3
    ns.clear()
    assert len(tmap) == 0
    ns.reload()
    assert len(tmap.initial) == 3
    assert len(tmap.cache) == 0
    assert len(e) == 0
    assert len(tmap) == 3 + len(e)
    assert cnt == 3


def test_SubProduct(claz):
    from fdi.dataset.testproducts import SP
    from fdi.dataset.baseproduct import BaseProduct

    Classes = claz

    # equivalence
    v = Classes.mapping['BaseProduct']
    assert v is BaseProduct

    y = SP()

    # register it in Classes so deserializer knows how to instanciate.
    Classes.update({'SP': SP})

    assert issubclass(SP, BaseProduct)

    from fdi.pal.context import MapContext

    class SSP(SP, MapContext):
        def __init__(self, **kwds):
            super().__init__(**kwds)

    x = SSP()
    x.instrument = 'ff'
    assert x.instrument == 'ff'
    x.rr = 'r'
    assert x.rr == 'r'


def test_Classes(claz):
    from fdi.dataset.baseproduct import BaseProduct
    from fdi.dataset.classes import All_Globals_Builtins
    PC = claz
    prjcls = PC.mapping
    nc = len(prjcls)
    assert nc > 44
    assert issubclass(prjcls.__class__, ChainMap)
    assert 'Product' in prjcls
    PC.clear()
    m = prjcls
    assert len(m) == len(All_Globals_Builtins)
    # clear() replenishes `initial` map
    PC.reload()
    assert len(prjcls) == len(prjcls.sources) + len(All_Globals_Builtins)
    c = prjcls['Product']
    assert c
    assert issubclass(c, BaseProduct)
    assert c(description='foobar').description == 'foobar'
    # Add a class to current namespace
    PC.update({'foo': int})
    # it shows in mapping
    assert 'foo' in prjcls
    # but not in PC.mapping.initial
    assert 'foo' not in prjcls.initial
    assert 'foo' not in prjcls.sources
    # mockup "permanent" updating PC.sources
    prjcls.sources.update({'path': 'sys'})
    assert 'path' in prjcls.sources
    assert 'path' not in prjcls
    # to make the change into effect, one has to re-run loading module_class
    PC.reload()
    assert 'path' in prjcls
    assert issubclass(prjcls['path'].__class__, list)
    # if there arre a lot of new classes e.g. there is a new package, to add
    # mockup permanent updating PC.sources with a dict of classname:class_obj/module_name

    class foo():
        pass

    PC.update({'maxsize': 'sys', 'f_': foo}, extension=True)
    assert 'maxsize' in prjcls.sources
    assert 'f_' in prjcls.sources
    assert 'maxsize' not in prjcls
    assert 'f_' not in prjcls
    # to make the change into effect, one has to re-run loading module_class
    PC.reload()
    assert 'maxsize' in prjcls
    assert prjcls['maxsize'] == sys.maxsize
    assert 'f_' in prjcls
    assert issubclass(prjcls['f_'], foo)

# add name spacw


def test_gb(claz):
    from fdi.dataset.baseproduct import BaseProduct

    Classes = claz
    Class_Look_Up = Classes.mapping
    nm = len(Classes.mapping.maps)
    assert 'wakaka' not in Class_Look_Up
    _bltn = dict((k, v) for k, v in {'wakaka': tuple}.items() if k[0] != '_')
    Classes.mapping.add_ns(_bltn, order=-1)
    assert nm+1 == len(Classes.mapping.maps)
    p = Class_Look_Up['Product']

    assert isinstance(p, type)
    assert issubclass(p, BaseProduct)
    assert 'Product' in Class_Look_Up.maps[0]
    assert Class_Look_Up.cache['Product'] is p

    assert 'wakaka' in Class_Look_Up.maps[-1]
    assert 'wakaka' in Class_Look_Up
    Classes.mapping.add_ns({'foo': bytes}, order=0)
    assert nm+2 == len(Classes.mapping.maps)
    assert 'foo' in Class_Look_Up.maps[0]
    assert 'foo' in Class_Look_Up



# -*- coding: utf-8 -*-

import logging

logger = logging.getLogger(__name__)


class Annotatable():
    """ An Annotatable object is an object that can give a
    human readable description of itself.
    """

    def __init__(self, description='UNKNOWN', **kwds):
        """

        Parameters
        ----------

        Returns
        -------

        """

        #print(__name__ + str(kwds))
        super().__init__(**kwds)  # Annotatable
        # skip if has been set by e.g. Attributable
        # if not hasattr(self, 'description'):
        self.description = description

    # @property
    # def description(self):
    #     """ xx must be a property for ``self.xx = yy`` to work in super class after xx is set as a property also by a subclass.

    #     Parameters
    #     ----------

    #     Returns
    #     -------

    #     """
    #     try:
    #         return self._description
    #     except AttributeError:
    #         return None

    # @description.setter
    # def description(self, description):
    #     """ Property of the description of this Annotatable object.

    #     Parameters
    #     ----------
    #     description : string
    #                      The new description.

    #     Returns
    #     -------

    #     """

    #     self._description = description

    def getDescription(self):
        """ gets the description of this Annotatable object.

        Parameters
        ----------

        Returns
        -------
        string:
              The current description.
        """
        return self.description

    def setDescription(self, newDescription):
        """ sets the description of this Annotatable object.

        Parameters
        ----------
        newDescription : string
                         The new description.

        Returns
        -------

        """
        self.description = newDescription




from txrequests import Session
from twisted.internet import defer

import pprint
import logging


def setuplogging():
    import logging
    import logging.config
    from . import logdict

    # create logger
    logging.config.dictConfig(logdict.logdict)
    logging.getLogger("requests").setLevel(logging.WARN)
    logging.getLogger("urllib3").setLevel(logging.WARN)
    logging.getLogger("filelock").setLevel(logging.WARN)
    return logging


logging = setuplogging()
logger = logging.getLogger()

logger.setLevel(logging.INFO)
logger.debug('logging level %d' % (logger.getEffectiveLevel()))


def clean_board():
    importlib.invalidate_caches()
    # importlib.reload(Classes)
    from fdi.dataset.classes import Classes
    return Classes.mapping


def pc():
    """ get configuration.

    """
    from fdi.utils.getconfig import getConfig as getc
    pns = getc(force=True)
    logger.debug(json.dumps(pns))
    return pns


def new_user_read_write(pc=pc()):
    """
    GIVEN a User model
    https://www.patricksoftwareblog.com/testing-a-flask-application-using-pytest/
    """
    new_user = User(pc['username'], pc['password'], roles='read_write')
    headers = auth_headers(pc['username'], pc['password'])
    return new_user, headers


def server(live_or_mock=live_or_mock(), new_user_read_write=new_user_read_write()):
    """ Server data from r/w user, mock or alive.

    """
    aburl, ty = live_or_mock
    user, headers = new_user_read_write
    headers['server_type'] = ty
    yield aburl, headers


def tmp_remote_storage(server=server(), client=client(), auth=auth()):
    """ temporary servered pool with module scope """
    aburl, headers = server
    poolid = 'test_remote_pool'
    pool = PoolManager.getPool(
        poolid, aburl + '/' + poolid, auth=auth, client=client)
    pool.removeAll()
    ps = ProductStorage(pool, client=client, auth=auth)
    assert issubclass(ps.getPool(poolid).client.__class__,
                      (requests.Session, FlaskClient))
    return ps


ps = tmp_remote_storage()
aburl, header = server()
pool = ps.getPool(ps.getPools()[0])
poolurl = pool.poolurl

Number = 10
refs = []


@defer.inlineCallbacks
def main():
    with Session(maxthreads=10) as session:
        def bg_cb(sess, resp):
            # parse the json storing the result on the response object
            resp.data = resp.json()
            return resp

        d = session.get(aburl,
                        background_callback=bg_cb)
        # do some other stuff, send some more requests while this one works
        response = yield d
        print('response status {0}'.format(response.status_code))
        # data will have been attached to the response object in the background
        pprint.pprint(response.data)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

from fdi.utils.jsonpath import jsonPath, flatten_compact
from fdi.dataset.schemas import getValidator, makeSchemaStore, validateJson
from fdi.utils.common import lls
import fdi.dataset.serializable
from fdi.dataset.serializable import serialize
from fdi.dataset.finetime import FineTime, FineTime1
from fdi.dataset.metadata import Parameter, MetaData, make_jsonable, guess_value
from fdi.dataset.listener import ListenerSet
from fdi.dataset.numericparameter import NumericParameter, BooleanParameter
from fdi.dataset.stringparameter import StringParameter
from fdi.dataset.dateparameter import DateParameter
from fdi.dataset.listener import EventSender, EventTypes, EventType, EventTypeOf, MetaDataListener, EventListener
from fdi.dataset.arraydataset import ArrayDataset, Column
from fdi.dataset.mediawrapper import MediaWrapper
from fdi.dataset.tabledataset import TableDataset
from fdi.dataset.dataset import Dataset, CompositeDataset
from fdi.dataset.unstructureddataset import UnstructuredDataset
from fdi.dataset.history import History
from fdi.dataset.baseproduct import BaseProduct
from fdi.dataset.product import Product
from fdi.dataset.datatypes import Vector, Vector2D, Quaternion
from fdi.dataset.classes import Classes
from fdi.dataset.serializable import serialize, Serializable
from fdi.pal.context import MapContext
from fdi.pal.productref import ProductRef
if 0:
    from fdi.dataset.annotatable import Annotatable
    from fdi.dataset.copyable import Copyable
    from fdi.dataset.odict import ODict
    from fdi.dataset.eq import deepcmp
    from fdi.dataset.deserialize import deserialize
    from fdi.dataset.quantifiable import Quantifiable
    from fdi.dataset.messagequeue import MqttRelayListener, MqttRelaySender
    from fdi.dataset.composite import Composite
    from fdi.dataset.metadataholder import MetaDataHolder
    from fdi.dataset.datatypes import DataTypes, DataTypeNames
    from fdi.dataset.attributable import Attributable
    from fdi.dataset.abstractcomposite import AbstractComposite
    from fdi.dataset.datawrapper import DataWrapper, DataWrapperMapper
    from fdi.dataset.indexed import Indexed
    from fdi.dataset.ndprint import ndprint
    from fdi.dataset.invalid import INVALID
    from fdi.dataset.browseproduct import BrowseProduct
    from fdi.dataset.readonlydict import ReadOnlyDict

from fdi.dataset.testproducts import SP, get_demo_product

from test_dataset import demo_TableDataset, demo_CompositeDataset

# from jsonschema import Draft7Validator as the_validator
from jsonschema import Draft201909Validator as the_validator
from jsonschema import RefResolver
from jsonschema.exceptions import RefResolutionError
from jsonschema import validate, ValidationError, SchemaError


import pytest
from pprint import pprint, pformat
import json
import copy
import array
import datetime
from datetime import timezone
import os
import os.path as op

import logging
# create logger
logger = logging.getLogger(__name__)
# logger.debug('level %d' %  (logger.getEffectiveLevel()))


SCHEMA_TEST_DATA = os.path.join(os.path.abspath(
    os.path.dirname(__file__)), 'resources/schema')
SCHEMA_DIR = os.path.abspath(os.path.join(
    os.path.dirname(__file__), '../fdi/schemas'))
SSP = 'https://fdi.schemas/%s%s'
""" Schema store prefix"""

verbose = 1 #False


@pytest.fixture(scope='function')
def schema_store():
    store = makeSchemaStore()
    return store


def check_examples_defaults(vtr, schema, jsn, paths):
    """check if "examples", "default" are valid."""

    if verbose:
        print(schema['title'])
        print('\n'.join(sorted(jsonPath(schema, '$..*', val='paths'))))

    for path in paths:
        for theproperty_name, theproperty in jsonPath(schema, path, val='full'):
            # theproperty_name is like 'allOf/1/properties/default/examples'
            mod = copy.deepcopy(jsn)
            if issubclass(theproperty.__class__, list):
                thelist = theproperty
            else:
                thelist = [theproperty]
            for n, example in enumerate(thelist):
                if verbose:
                    print('Auto Validating %s["%s"][%d] %s...' % (
                        schema['title'], theproperty_name, n, example))
                if 'properties/value' in theproperty_name:
                    mod['value'] = example
                    assert vtr.validate(mod) is None
                elif 'properties/examples' in theproperty_name or \
                        'examples' == theproperty_name:
                    mod = example
                    assert vtr.validate(mod) is None
                else:
                    prop = theproperty_name.split(
                        'properties/', 1)[-1].split('/', 1)[0]
                    if prop == 'default' and 'value' in jsn:
                        # default ,means value default
                        mod['value'] = example
                    else:
                        mod[prop] = example
                    assert vtr.validate(mod) is None


def check_general(vtr, jsn, name):
    # STID
    assert vtr.validate(jsn) is None
    # change _STID
    bad = copy.deepcopy(jsn)
    bad['_STID'] = bad['_STID'][:-1]
    with pytest.raises(ValidationError):
        assert vtr.validate(bad) is None
    # remove _STID
    bad = copy.deepcopy(jsn)
    st = bad.pop('_STID', '')
    assert name in st
    with pytest.raises(ValidationError):
        assert vtr.validate(bad) is None
    bad = copy.deepcopy(jsn)
    # remove metadata
    if bad.pop('meta', 0) or bad.pop('_ATTR_meta', 0):
        with pytest.raises(ValidationError):
            assert vtr.validate(bad) is None

    if vtr.is_type(vtr.schema, 'object'):
        # change Require_STID to false
        sch = copy.deepcopy(vtr.schema)
        sch['$ref'] = "https://fdi.schemas/Require_STID_False"
        vtr = getValidator(sch, verbose=verbose)
        bad = copy.deepcopy(jsn)
        st = bad.pop('_STID', '')
        # STID is not required.
        assert vtr.validate(bad) is None


def test_validator():
    #fdi_sch_dir = op.dirname(op.join(op.abspath(__file__), '../fdi/schemas'))
    sch_dir = op.join(op.abspath(op.dirname(__file__)), 'resources/schemas')
    sch_path = op.join(sch_dir, 'prd_schema.jsn')
    with open(sch_path, 'r') as file:
        sch = json.load(file)

    from fdi.dataset.baseproduct import BaseProduct
    jsn = json.loads(BaseProduct().serialized())

    store = makeSchemaStore(sch_dir)
    assert getValidator(sch, schema_store=store, base_schema=sch,
                        verbose=1).validate(jsn) is None


def test_true_false(schema_store):
    # load schemas
    sch_True = schema_store[SSP % ('', 'True')]
    sch_False = schema_store[SSP % ('', 'False')]
    # "$ref": "True"  "foo": {"type": "string"} "bar": {"type": "string"}
    use_true = json.loads("""{
        "$id": "https://fdi.schemas/use_true",
        "if": {"$ref": "https://fdi.schemas/True"},
        "then": {
            "required": ["foo"]
        }
    }""")

    # pass
    vtr = getValidator(use_true, verbose=verbose)
    assert not validateJson({"foo": "k", "a": 1}, vtr)
    assert vtr.validate({"foo": "k"}) is None
    # error
    assert validateJson({"food": "k"}, vtr)
    with pytest.raises(ValidationError):
        assert vtr.validate({"food": "k", "a": 2}) is None
    with pytest.raises(ValidationError):
        assert vtr.validate({}) is None

    use_false = json.loads("""{
        "$id": "https://fdi.schemas/use_true",
        "if": {"$ref": "False"},
        "then": {
            "required": ["foo"]
        }
    }""")

    if 0:
        also_run = {
            "$id": "https://fdi.schemas/use_false",
            # "propertyNames": False,
            # "foo": "string",
            "if":
            False,
            "then": {
                "required": ["foo"]
            }
        }

    vtr = getValidator(use_false, verbose=verbose)
    # pass
    assert not validateJson({"foo": "k", "a": 1}, vtr)
    assert vtr.validate({"foo": "k"}) is None

    assert not validateJson({"food": "k"}, vtr)
    assert vtr.validate({"food": "k", "a": 2}) is None
    assert vtr.validate({}) is None


def test_bytes(schema_store):
    scn = 'bytes'
    sch = schema_store[SSP % ('', scn)]
    byts = b'0123456789\x0a\x0b\x0c\x0d\x0e\x0f\x10\xff'
    saved = fdi.dataset.serializable.GZIP
    vtr = getValidator(sch, verbose=verbose)
    try:
        for gzip in [True, False]:
            fdi.dataset.serializable.GZIP = gzip
            for typ in (bytes, bytearray):
                jsn = json.loads(serialize(typ(byts)))
                logger.info(jsn)
                assert vtr.validate(jsn) is None

            if gzip:
                # no valid STID
                bad = {
                    'code': 'H4sIABqX\n', '_S': 'bytes,gz,b64'}
                with pytest.raises(ValidationError):
                    assert vtr.validate(bad) is None
                # invalid _STID
                bad = {
                    'code': 'H4sIABqX\n', '_STID': 'bytes,gz,b64g'}
                with pytest.raises(ValidationError):
                    assert vtr.validate(bad) is None
                # invalid characters not checked
                bad = {'code': 'H4sI!@qX\n', '_STID': 'bytes,gz,b64'}
                # with pytest.raises(ValidationError):
                assert vtr.validate(bad) is None
            else:
                # invalid characters not checked
                bad = {'code': '3031323334350g', '_STID': 'bytes'}
                # with pytest.raises(ValidationError):
                assert vtr.validate(bad) is None
    finally:
        fdi.dataset.serializable.GZIP = saved

    # "example ignores jsn
    check_examples_defaults(vtr, sch, jsn, [
        'examples',
    ])


def test_a_array(schema_store):
    scn = 'a_array'
    sch = schema_store[SSP % ('', scn)]
    byts = b'0123456789\x0a\x0b\x0c\x0d\x0e\x0f'
    vtr = getValidator(sch, verbose=verbose)

    saved = fdi.dataset.serializable.GZIP
    try:
        for gzip in [True, False]:
            fdi.dataset.serializable.GZIP = gzip
            for tcode in ('I', 'H'):
                jsn = json.loads(serialize(array.array(tcode, byts)))
                logger.info(jsn)
                assert vtr.validate(jsn) is None

            if not gzip:
                # bad STID
                bad = {'_STID': '',
                       'code': '303132333435363738390a0b0c0d0e0f'}
                with pytest.raises(ValidationError):
                    assert vtr.validate(bad) is None
    finally:
        fdi.dataset.serializable.GZIP = saved

    # "example ignores jsn
    check_examples_defaults(vtr, sch, jsn, ['examples'])


def test_RQSTID(schema_store):
    """ Test setting requires:['_STID'] with schema "Require_STID".
    """
    scn = 'TEST_STID'
    sch = schema_store[SSP % ('', scn)]

    sch['$ref'] = "Require_STID_True"
    vtr = getValidator(sch, verbose=verbose)
    jsn = {"foo": "bar", "_STID": "byte"}
    assert vtr.validate(jsn) is None
    # error from no _STID
    jsn = {"foo": "bar", "m_STID": "byte"}
    with pytest.raises(ValidationError):
        assert vtr.validate(jsn) is None
    with pytest.raises(ValidationError):
        assert vtr.validate({}) is None

    # pass
    sch['$ref'] = "Require_STID_False"
    vtr = getValidator(sch, verbose=verbose)
    jsn = {"foo": "bar", "_STID": "byte"}
    assert vtr.validate(jsn) is None
    jsn = {"foo": "bar", "m_STID": "byte"}
    assert vtr.validate(jsn) is None
    assert vtr.validate({}) is None


def test_preset_STID(schema_store):

    scn = 'TEST_STID'
    sch = schema_store[SSP % ('', scn)]

    # constant preset _STID
    sch['properties']['_STID'] = {"const": "Foo"}
    vtr = getValidator(sch, verbose=verbose)
    jsn = {"foo": "bar", "_STID": "Foo"}
    assert vtr.validate(jsn) is None
    # wrong STID
    jsn = {"foo": "bar", "_STID": "bytes"}
    with pytest.raises(ValidationError):
        assert vtr.validate(jsn) is None
    # missing
    jsn = {"foo": "bar"}
    with pytest.raises(ValidationError):
        assert vtr.validate(jsn) is None
    # also wrong
    jsn = json.loads('{"foo":"bar", "_STID":"bytes,gz"}')
    with pytest.raises(ValidationError):
        assert vtr.validate(jsn) is None
    # empty
    with pytest.raises(ValidationError):
        assert vtr.validate({}) is None

    # modift schema to have list of possible IDs in preset _STID
    sch['properties']['_STID'] = {"enum": ["bytes", "bytes,gz"]}
    vtr = getValidator(sch, verbose=verbose)
    jsn = {"foo": "bar", "_STID": "bytes,gz"}
    assert vtr.validate(jsn) is None
    jsn = {"foo": "bar", "_STID": "bytes"}
    assert vtr.validate(jsn) is None

    # errors in STID in instane
    jsn = {"foo": "bar", "_STID": "bytes,g"}
    with pytest.raises(ValidationError):
        assert vtr.validate(jsn) is None
    with pytest.raises(ValidationError):
        assert vtr.validate({"_STID": None}) is None

    # Properties is anyOf
    del sch['properties']
    sch["anyOf"] = [
        {'properties': {
            "foo": {"const": "deadbeef"},
            "_STID": {"type": "string", "pattern": "^bytes$"}
        }},
        {'properties': {
            "_STID": {"type": "string", "pattern": "bytes.gz$"}}
         }
    ]

    vtr = getValidator(sch, verbose=verbose)
    jsn = {"foo": "bar", "_STID": "bytes,gz"}
    assert vtr.validate(jsn) is None
    jsn = {"foo": "deadbeef", "_STID": "bytes"}
    assert vtr.validate(jsn) is None

    # error in foo: when STID=='gzip'
    jsn = {"foo": "bar", "_STID": "bytes"}
    with pytest.raises(ValidationError):
        assert vtr.validate(jsn) is None

    # errors in STID in instane
    jsn = {"foo": "bar", "_STID": "bytes,g"}
    with pytest.raises(ValidationError):
        assert vtr.validate(jsn) is None
    with pytest.raises(ValidationError):
        assert vtr.validate({"_STID": None}) is None

    # modify STID to ""
    jsn = {"foo": "bar", "_STID": "g"}
    with pytest.raises(ValidationError):
        assert vtr.validate(jsn) is None


def test_urn(schema_store):
    scn = 'Urn'
    sch = schema_store[SSP % ('pal/', scn)]

    if 1:
        jsn = json.loads(serialize('urn:pool:si.PL:20'))
    vtr = getValidator(sch, verbose=verbose)
    assert vtr.validate(jsn) is None

    check_examples_defaults(vtr, sch, jsn, [
                            'examples',
                            ])

    # invalid urn in 'urns'
    cpy = copy.deepcopy(jsn)
    bad = cpy.replace("urn:", "urn::")
    with pytest.raises(ValidationError):
        assert getValidator(sch, verbose=verbose).validate(bad) is None


def test_STID(schema_store):
    scn = 'STID'
    sch = schema_store[SSP % ('', scn)]

    jsn = json.loads(serialize('FineTime'))
    vtr = getValidator(sch, verbose=verbose)
    assert vtr.validate(jsn) is None

    check_examples_defaults(vtr, sch, jsn, [
                            'examples',
                            ])

    # invalid
    cpy = copy.deepcopy(jsn)
    bad = cpy.replace("F", "#")
    with pytest.raises(ValidationError):
        assert vtr.validate(bad) is None
    with pytest.raises(ValidationError):
        assert vtr.validate(None) is None

    # extended


def test_datetime(schema_store):
    scn = 'datetime'
    sch = schema_store[SSP % ('', scn)]

    then = datetime.datetime(
        2019, 2, 19, 1, 2, 3, 456789, tzinfo=timezone.utc)
    jsn = json.loads(serialize(then, indent=4))
    if verbose:
        logger.info(jsn)

    vtr = getValidator(sch, verbose=verbose)
    assert vtr.validate(jsn) is None

    check_general(vtr, jsn, scn)

    # can be None
    assert vtr.validate(None) is None


def test_ListenerSet(schema_store):
    scn = 'ListenerSet'
    sch = schema_store[SSP % ('dataset/', scn)]

    jsn = json.loads(serialize(ListenerSet(), indent=4))
    if verbose:
        logger.info(jsn)

    vtr = getValidator(sch, verbose=verbose)
    assert vtr.validate(jsn) is None

    check_general(vtr, jsn, scn)

    # can be None
    assert vtr.validate(None) is None


def test_FineTime(schema_store):
    scn = 'FineTime'
    sch = schema_store[SSP % ('dataset/', scn)]

    if 1:
        jsn = json.loads(serialize(FineTime(123456789876543)))
    else:
        jsn_path = os.path.join(SCHEMA_TEST_DATA, scn0 + scn + '.jsn')
        with open(jsn_path, 'r') as file:
            jsn = json.load(file)
    # print(jsn)
    vtr = getValidator(sch, verbose=verbose)
    check_general(vtr, jsn, scn)
    check_examples_defaults(vtr, sch, jsn, [
                            'properties.format.examples',
                            'examples',
                            ])

    assert vtr.validate(jsn) is None
    # invalid
    bad = copy.deepcopy(jsn)
    bad['tai'] = 9.9
    with pytest.raises(ValidationError):
        assert vtr.validate(bad) is None
    bad = copy.deepcopy(jsn)
    bad['format'] = 'd'
    with pytest.raises(ValidationError):
        assert vtr.validate(bad) is None


def test_xParameter(schema_store):
    scn = 'Parameter'
    sch = schema_store[SSP % ('dataset/', scn)]
    x = Parameter(3.14,
                  description="foo",
                  default=42,
                  valid={3.14: 'ok'},
                  typ_=None)
    jsn = json.loads(serialize(x, indent=4))

    if verbose:
        logger.info(jsn)
    assert jsn['type'] == 'float'
    vtr = getValidator(sch, verbose=verbose)
    assert vtr.validate(jsn) is None

    check_general(vtr, jsn, scn)
    check_examples_defaults(vtr, sch, jsn, [
                            'allOf.[1].properties.default.examples',
                            'allOf.[1].properties.default.default',
                            'allOf.[1].properties.value.examples',
                            'examples',
                            ])

    assert jsn['valid'] is not None
    jsn['valid'] = None
    assert jsn['valid'] is None
    vtr.validate(jsn) is None
    jsn['description'] = None
    vtr.validate(jsn) is None

    # invalid
    bad = copy.deepcopy(jsn)
    del bad['value']
    with pytest.raises(ValidationError):
        assert vtr.validate(bad) is None
    bad = copy.deepcopy(jsn)
    bad['valid'] = 4
    with pytest.raises(ValidationError):
        assert vtr.validate(bad) is None

    bad = copy.deepcopy(jsn)
    bad['type'] = 'foo'
    with pytest.raises(ValidationError):
        k = vtr.validate(bad)
    bad = copy.deepcopy(jsn)
    bad['valid'] = [[1, 2]]
    vtr.validate(bad) is None
    bad['valid'] = [[1], [1, 2]]
    with pytest.raises(ValidationError):
        vtr.validate(bad) is None
    # logger.warning(str(vtr.validate(bad)))


def test_BooleanParameter(schema_store):
    scn = 'BooleanParameter'
    sch = schema_store[SSP % ('dataset/', scn)]

    if 1:
        jsn = json.loads(serialize(BooleanParameter(True,
                                                    description="foo?",
                                                    default=None,
                                                    ),
                                   indent=4))
    if verbose:
        logger.info(jsn)
    assert jsn['type'] == 'boolean'
    assert jsn['default'] is None
    vtr = getValidator(sch, verbose=verbose)
    assert vtr.validate(jsn) is None

    check_general(vtr, jsn, scn)
    check_examples_defaults(vtr, sch, jsn, [
                            'allOf.[1].properties.value.examples',
                            'allOf.[1].properties.value.default',
                            'allOf.[1].properties.default.examples',
                            'allOf.[1].properties.default.default',
                            'examples',
                            ])

    # can be None?
    with pytest.raises(ValidationError):
        assert vtr.validate(None) is None

    # invalid
    bad = copy.deepcopy(jsn)
    bad['type'] = 3
    bad['value'] = 'd'
    with pytest.raises(ValidationError):
        assert vtr.validate(bad) is None


def test_StringParameter(schema_store):
    scn = 'StringParameter'
    sch = schema_store[SSP % ('dataset/', scn)]

    if 1:
        jsn = json.loads(serialize(StringParameter("wahah",
                                                   description="tester",
                                                   typecode='10B'
                                                   ),
                                   indent=4)
                         )
    if verbose:
        logger.info(jsn)
    if 'type' in jsn:
        assert jsn['type'] == 'string'
    assert jsn['default'] is None
    vtr = getValidator(sch, verbose=verbose)
    assert vtr.validate(jsn) is None

    check_general(vtr, jsn, scn)
    check_examples_defaults(vtr, sch, jsn, [
                            'allOf.[1].properties.value.examples',
                            'allOf.[1].properties.value.default',
                            'allOf.[1].properties.default.examples',
                            'allOf.[1].properties.default.default',
                            'allOf.[1].properties.typecode.examples',
                            'examples',
                            ])

    # can be None?
    with pytest.raises(ValidationError):
        assert vtr.validate(None) is None

    # invalid
    bad = copy.deepcopy(jsn)
    bad['type'] = 3
    bad['value'] = 'd'
    with pytest.raises(ValidationError):
        assert vtr.validate(bad) is None


def test_NumericParameter(schema_store):
    scn = 'NumericParameter'
    sch = schema_store[SSP % ('dataset/', scn)]

    if 1:
        jsn = json.loads(serialize(NumericParameter(3.14,
                                                    description="foo",
                                                    default=42,
                                                    valid={0: 'ok'},
                                                    typ_=None,
                                                    unit='lyr'),
                                   indent=4)
                         )
    if verbose:
        logger.info(jsn)
    assert jsn['unit'] == 'lyr'
    assert jsn['typecode'] == None
    vtr = getValidator(sch, verbose=verbose)

    check_general(vtr, jsn, scn)
    check_examples_defaults(vtr, sch, jsn, [
                            'properties.default.examples',
                            'properties.default.default',
                            'properties.value.examples',
                            'examples',
                            ])

    # can not be None.
    with pytest.raises(ValidationError):
        assert vtr.validate(None) is None
    assert vtr.validate(jsn) is None

    # invalid
    bad = copy.deepcopy(jsn)
    bad['typecode'] = 3
    bad['value'] = 'd'
    with pytest.raises(ValidationError):
        assert vtr.validate(bad) is None

    v = NumericParameter(
        value=[0b01], description='valid rules described with binary masks',
        typ_='list',
        default=[0b00],
        valid={(0b011000, 0b01): 'on', (0b011000, 0b00): 'off'},
        typecode='H')

    jsn = json.loads(serialize(v))
    vtr = getValidator(sch, verbose=verbose)
    assert vtr.validate(jsn) is None

    # Parameter won't cut it
    sch_para = schema_store[SSP % ('dataset/', 'Parameter')]
    vtr_para = getValidator(sch_para, verbose=verbose)
    with pytest.raises(ValidationError):
        assert vtr_para.validate(jsn) is None
    x_para = Parameter(3.14,
                       description="foo",
                       default=42,
                       valid={3.14: 'ok'},
                       typ_=None)
    jsn_para = json.loads(serialize(x_para, indent=4))
    with pytest.raises(ValidationError):
        assert vtr.validate(jsn_para) is None

    # Vector


def test_Vector(schema_store):
    scn = 'Vector'
    sch = schema_store[SSP % ('dataset/', scn)]

    if 1:
        jsn = json.loads(serialize(Vector((3.14, 5, 0xb),
                                          typecode='f',
                                          unit='au'),
                                   indent=4)
                         )
    if verbose:
        logger.info(jsn)
    assert jsn['unit'] == 'au'
    assert jsn['typecode'] == 'f'
    vtr = getValidator(sch, verbose=verbose)

    check_general(vtr, jsn, scn)
    check_examples_defaults(vtr, sch, jsn, [
                            'properties.default.examples',
                            'properties.default.default',
                            'properties.value.examples',
                            'examples',
                            ])

    # can not be None.
    with pytest.raises(ValidationError):
        assert vtr.validate(None) is None
    assert vtr.validate(jsn) is None

    # invalid
    bad = copy.deepcopy(jsn)
    bad['typecode'] = 3
    bad['value'] = 'd'
    with pytest.raises(ValidationError):
        assert vtr.validate(bad) is None


def test_DateParameter(schema_store):
    scn = 'DateParameter'
    sch = schema_store[SSP % ('dataset/', scn)]
    then = datetime.datetime(
        2019, 2, 19, 1, 2, 3, 456789, tzinfo=timezone.utc)
    v = DateParameter(value=FineTime(then), description='date param',
                      default=99,
                      valid={(0, 9876543210123456): 'xy'})
    tc = v.typecode

    jsn = json.loads(serialize(v))

    if verbose:
        logger.info("JSON instance to test: %s" % jsn)
    assert jsn['default']['tai'] == 99
    assert jsn['typecode'] == tc == "%Y-%m-%dT%H:%M:%S.%f"
    vtr = getValidator(sch, verbose=verbose)

    check_general(vtr, jsn, scn)
    check_examples_defaults(vtr, sch, jsn, [
                            'allOf.[1].properties.default.examples',
                            'allOf.[1].properties.default.default',
                            'allOf.[1].properties.value.examples',
                            'examples',
                            ])

    assert vtr.validate(jsn) is None
    # default can also be None and integer of TAI value.
    jsn['default'] = 0
    assert vtr.validate(jsn) is None
    # value is assigned with TAI
    v = DateParameter(value=1234, description='value=1234',
                      default=99,
                      valid={(0, 987): 'val'})
    # "value" attribute is still FineTime object
    assert issubclass(v.value.__class__, FineTime)
    assert v.value.tai == 1234
    jsn = json.loads(serialize(v))
    assert vtr.validate(jsn) is None

    # can not be None?
    with pytest.raises(ValidationError):
        assert vtr.validate(None) is None

    # invalid
    # value is not FineTime
    bad = copy.deepcopy(jsn)
    bad['value'] = {"foo": 4}
    with pytest.raises(ValidationError):
        assert vtr.validate(bad) is None
    bad = copy.deepcopy(jsn)
    bad['typecode'] = 3
    with pytest.raises(ValidationError):
        assert vtr.validate(bad) is None
    bad = copy.deepcopy(jsn)
    bad['value'] = 'd'
    with pytest.raises(ValidationError):
        assert vtr.validate(bad) is None


def test_MetaData(schema_store):
    scn = 'MetaData'
    sch = schema_store[SSP % ('dataset/', scn)]

    v = MetaData()
    v['par'] = Parameter(description='test param', value=534)
    a1 = 'a test NumericParameter'
    a2 = 100.234
    a3 = 'second'
    a4 = 'float'
    a5 = 0
    a6 = ''
    a7 = 'f'
    v['num'] = NumericParameter(description=a1, value=a2, unit=a3,
                                typ_=a4, default=a5, valid=a6, typecode=a7)
    then = datetime.datetime(
        2019, 2, 19, 1, 2, 3, 456789, tzinfo=timezone.utc)
    v['dat'] = DateParameter(value=FineTime(then), description='date param',
                             default=99,
                             valid={(0, 9876543210123456): 'xy'})

    a1 = 'a test BooleanParameter'
    a2 = 100.234
    a5 = True
    a6 = [[(True, False), "all"]]
    v['boo'] = BooleanParameter(description=a1, value=a2,
                                default=a5, valid=a6)
    a1 = 'a test StringcParameter'
    a2 = 'eeeee'
    a3 = 'second'
    a4 = 'string'
    a5 = ''
    a6 = '9B'
    v['str'] = StringParameter(description=a1, value=a2, default=a3,
                               valid=a5, typecode=a6)

    # listeners

    class MockMetaListener(MetaDataListener):
        pass
    lis1 = MockMetaListener('foo')
    lis2 = MockMetaListener('bar')

    v.addListener(lis1)
    v.addListener(lis2)

    jsn = json.loads(serialize(v))

    if 1 or verbose:
        logger.info("JSON instance to test: %s" %
                    lls(pformat(jsn, indent=4), 2000))
    # assert jsn['default']['tai'] == 99
    # assert jsn['typecode'] == tc == "%Y-%m-%dT%H:%M:%S.%f"
    vtr = getValidator(sch, verbose=verbose)
    assert vtr.validate(jsn) is None
    check_general(vtr, jsn, scn)


def test_Dataset(schema_store):
    scn = 'Dataset'
    sch = schema_store[SSP % ('dataset/', scn)]

    v = Dataset(description='test Dataset')
    v.meta = MetaData(description='test Dataset.MEtadata')
    v.data = 987.4

    jsn = json.loads(serialize(v))

    if 1 or verbose:
        logger.info("JSON instance to test: %s" %
                    lls(pformat(jsn, indent=4), 2000))
    # assert jsn['default']['tai'] == 99
    # assert jsn['typecode'] == tc == "%Y-%m-%dT%H:%M:%S.%f"
    vtr = getValidator(sch, verbose=verbose)
    assert vtr.validate(jsn) is None

    check_general(vtr, jsn, scn)


def test_ArrayDataset(schema_store):
    scn = 'ArrayDataset'
    sch = schema_store[SSP % ('dataset/', scn)]

    for atype in (list, bytes):

        if issubclass(atype([]).__class__, (bytes, bytearray)):
            a1 = atype([1, 44, 0xff])      # an array of data
            a4 = 'integer'              # type
            a6 = 'H'                  # typecode
        else:
            a1 = atype([1, 4.4, 5.4E3])      # an array of data
            a4 = 'float'              # type
            a6 = 'f'                  # typecode
        a2 = 'ev'                 # unit
        a3 = 'three energy vals'  # description
        a7 = (8, 9)

        v = ArrayDataset(data=a1, unit=a2, description=a3,
                         typ_=a4, typecode=a6)

        jsn = json.loads(serialize(v))

        if verbose:
            logger.info("JSON instance to test: %s" %
                        lls(pformat(jsn, indent=4), 2000))
        # assert jsn['default']['tai'] == 99
        # assert jsn['typecode'] == tc == "%Y-%m-%dT%H:%M:%S.%f"
        vtr = getValidator(sch, verbose=verbose)
        assert vtr.validate(jsn) is None

        check_general(vtr, jsn, scn)

    # subclasses


def test_Column(schema_store):

    scn = 'Column'
    sch = schema_store[SSP % ('dataset/', scn)]
    v = Column(data=[4, 9], unit='m')
    jsn = json.loads(serialize(v))

    if verbose:
        logger.info("JSON instance to test: %s" %
                    lls(pformat(jsn, indent=4), 2000))
    vtr = getValidator(sch, verbose=verbose)
    assert vtr.validate(jsn) is None

    check_general(vtr, jsn, scn)


im_svg = '<svg aria-hidden="true" data-prefix="far" data-icon="copy" class="svg-inline--fa fa-copy fa-w-14" role="img" xmlns="http://www.w3.org/2000/svg" viewBox="0 0 448 512"><path fill="#777" d="M433.941 65.941l-51.882-51.882A48 48 0 0 0 348.118 0H176c-26.51 0-48 21.49-48 48v48H48c-26.51 0-48 21.49-48 48v320c0 26.51 21.49 48 48 48h224c26.51 0 48-21.49 48-48v-48h80c26.51 0 48-21.49 48-48V99.882a48 48 0 0 0-14.059-33.941zM266 464H54a6 6 0 0 1-6-6V150a6 6 0 0 1 6-6h74v224c0 26.51 21.49 48 48 48h96v42a6 6 0 0 1-6 6zm128-96H182a6 6 0 0 1-6-6V54a6 6 0 0 1 6-6h106v88c0 13.255 10.745 24 24 24h88v202a6 6 0 0 1-6 6zm6-256h-64V48h9.632c1.591 0 3.117.632 4.243 1.757l48.368 48.368a6 6 0 0 1 1.757 4.243V112z"></path></svg>'


def test_MediaWrapper(schema_store):

    scn = 'MediaWrapper'
    sch = schema_store[SSP % ('dataset/', scn)]
    v = MediaWrapper(data=im_svg,
                     description='copybutton in svg',
                     typ_='image/svg'
                     )
    _s = serialize(v)

    jsn = json.loads(_s)

    if verbose:
        logger.info("JSON instance to test: %s" %
                    lls(pformat(jsn, indent=4), 2000))
    vtr = getValidator(sch, verbose=verbose)
    assert vtr.validate(jsn) is None

    check_general(vtr, jsn, scn)


def test_TableDataset(schema_store):
    scn = 'TableDataset'
    sch = schema_store[SSP % ('dataset/', scn)]

    v = demo_TableDataset()

    # v.meta = MetaData(description='test TableDataset.MEtadata')
    # v.data = 987.4

    jsn = json.loads(serialize(v))

    if 1 or verbose:
        logger.info("JSON instance to test: %s" %
                    lls(pformat(jsn, indent=4), 2000))
    # assert jsn['default']['tai'] == 99
    # assert jsn['typecode'] == tc == "%Y-%m-%dT%H:%M:%S.%f"
    vtr = getValidator(sch, verbose=verbose)
    assert vtr.validate(jsn) is None
    check_general(vtr, jsn, scn)


def test_CompositeDataset(schema_store):
    scn = 'CompositeDataset'
    sch = schema_store[SSP % ('dataset/', scn)]

    if 0:
        v = CompositeDataset(description="tester CompositeDataset",
                             data=[('col1', [1, 4.4, 5.4E3], 'eV'),
                                   ('col2', [0, 43.2, 2E3], 'cnt')
                                   ])
    v = demo_CompositeDataset()

    # v.meta = MetaData(description='test CompositeDataset.MEtadata')
    # v.data = 987.4

    jsn = json.loads(serialize(v))

    if 1 or verbose:
        logger.info("JSON instance to test: %s" %
                    lls(pformat(jsn, indent=4), 2000))
    # assert jsn['default']['tai'] == 99
    # assert jsn['typecode'] == tc == "%Y-%m-%dT%H:%M:%S.%f"
    vtr = getValidator(sch, verbose=verbose)
    assert vtr.validate(jsn) is None

    check_general(vtr, jsn, scn)


def test_UnstructuredDataset(schema_store):
    scn = 'UnstructuredDataset'
    sch = schema_store[SSP % ('dataset/', scn)]

    v = UnstructuredDataset()

    p = demo_CompositeDataset()
    v.put(p.serialized(), 'json')

    # v.meta = MetaData(description='test UnstructuredDataset.MEtadata')
    # v.data = 987.4

    jsn = json.loads(serialize(v))

    if 1 or verbose:
        logger.info("JSON instance to test: %s" %
                    lls(pformat(jsn, indent=4), 2000))
    # assert jsn['default']['tai'] == 99
    # assert jsn['typecode'] == tc == "%Y-%m-%dT%H:%M:%S.%f"
    vtr = getValidator(sch, verbose=verbose)
    assert vtr.validate(jsn) is None

    check_general(vtr, jsn, scn)


def test_History(schema_store):
    scn = 'History'
    sch = schema_store[SSP % ('dataset/', scn)]

    # p = get_demo_product()

    v = History()
    dt = datetime.datetime(
        2019, 2, 19, 1, 2, 3, 456789, tzinfo=timezone.utc)
    v.add_input(args={'id': 1, 'width': 2.33, 'name': 'asd',
                      'ok': True, 'speed': [4, 5],
                      'scores': array.array('f', [6, 7]),
                      'and': None, 'a': 'b', 'c': [11],
                      'dt': dt
                      },
                info=dict(c='d')
                )

    # odd ball
    parray = v.meta['scores']
    assert parray.value == array.array('f', [6, 7])
    sparray = json.loads(serialize(parray))
    pasc = schema_store[SSP % ('dataset/', 'Parameter')]
    tor = getValidator(pasc, verbose=verbose)
    res=tor.validate(sparray)
    assert res is None

    jsn = json.loads(serialize(v))

    if 1 or verbose:
        logger.info("JSON instance to test: %s" %
                    lls(pformat(jsn, indent=4), 2000))

    # assert jsn['default']['tai'] == 99
    # assert jsn['typecode'] == tc == "%Y-%m-%dT%H:%M:%S.%f"
    vtr = getValidator(sch, verbose=verbose)
    assert vtr.validate(jsn) is None

    check_general(vtr, jsn, scn)


def test_BaseProduct(schema_store):
    scn = 'BaseProduct'
    sch = schema_store[SSP % ('dataset/', scn)]

    # p = get_demo_product()

    v = BaseProduct(description='tester Base')
    i0 = 6
    i1 = [[1, 2, 3], [4, 5, i0], [7, 8, 9]]
    i2 = 'ev'                 # unit
    i3 = 'img1'  # description
    image = ArrayDataset(data=i1, unit=i2, description=i3)

    v["RawImage"] = image
    v.set('QualityImage', ArrayDataset(
        [[0.1, 0.5, 0.7], [4e3, 6e7, 8], [-2, 0, 3.1]]))
    # add a tabledataset
    s1 = [('col1', [1, 4.4, 5.4E3], 'eV'),
          ('col2', [0, 43.2, 2E3], 'cnt')
          ]
    spec = TableDataset(data=s1)
    v["Spectrum"] = spec

    # v.meta = MetaData(description='test BaseProduct.MEtadata')
    # v.data = 987.4

    jsn = json.loads(serialize(v))

    if 1 or verbose:
        logger.info("JSON instance to test: %s" %
                    lls(pformat(jsn, indent=4), 2000))
    # assert jsn['default']['tai'] == 99
    # assert jsn['typecode'] == tc == "%Y-%m-%dT%H:%M:%S.%f"
    vtr = getValidator(sch, verbose=verbose)
    assert vtr.validate(jsn) is None

    check_general(vtr, jsn, scn)


def test_MapContext(schema_store):
    scn = 'MapContext'
    sch = schema_store[SSP % ('dataset/', scn)]

    v = MapContext(description='MapContext demo')
    r = ProductRef(v)
    v.refs['self'] = r
    v['arr']=ArrayDataset([6,4,2])

    jsn = json.loads(serialize(v))

    if verbose:
        logger.info("JSON instance to test: %s" %
                    lls(pformat(jsn, indent=4), 2000))
    # assert jsn['default']['tai'] == 99
    # assert jsn['typecode'] == tc == "%Y-%m-%dT%H:%M:%S.%f"
    vtr = getValidator(sch, verbose=verbose)
    assert vtr.validate(jsn) is None

    check_general(vtr, jsn, scn)


def XXXtest_all(schema_store):

    all_cls = Classes.mapping
    all_schemas = list(schema_store)
    validator_list = {}
    for c_name, c_class in all_cls.items():
        if c_name == 'Serializable':
            continue
        if isinstance(c_class, type) and issubclass(c_class, Serializable):
            jsn = json.loads(serialize(c_class()))
            for sch_name, sch in schema_store.items():
                if c_name in sch_name:
                    logger.info('Found class %s for schema %s.' %
                                (c_name, sch_name))
                    if sch_name not in validator_list:
                        validator_list[sch_name] = getValidator(
                            sch, verbose=verbose)
                    assert validator_list[sch_name].validate(jsn) is None

    sch = schema_store[SSP % ('dataset/', scn)]

    v = get_demo_product()

    # v.meta = MetaData(description='test MapContext.MEtadata')
    # v.data = 987.4

    jsn = json.loads(serialize(v))


def te():
    try:
        assert validate(jsn, sch)
    except SchemaError as e:
        print("There is an error with the schema")
        assert False
    except ValidationError as e:
        print(e)

        print("---------")
        print(e.absolute_path)

        print("---------")
        print(e.absolute_schema_path)
        assert False



# -*- coding: utf-8 -*-

from ..utils.common import mstr, bstr
from .odict import ODict
from .listener import EventListener
from .datawrapper import DataWrapperMapper
from .composite import Composite
from .annotatable import Annotatable
from .attributable import Attributable

import logging
# create logger
logger = logging.getLogger(__name__)
# logger.debug('level %d' %  (logger.getEffectiveLevel()))


class AbstractComposite(Attributable, EventListener, Composite, DataWrapperMapper):
    """ an annotatable and attributable subclass of Composite.

    Composite inherits annotatable via EventListener via DataContainer.
    """

    def __init__(self, **kwds):
        """

        Parameters
        ----------

        Returns
        -----

        """

        super().__init__(**kwds)

    def toString(self, level=0, extra=False, param_widths=None,
                 tablefmt='grid', tablefmt1='simple', tablefmt2='psql',
                 width=0,
                 matprint=None, trans=True, beforedata='', heavy=True,
                 center=-1, **kwds):
        """ matprint: an external matrix print function

        Parameters
        ----------
        level : int
            Detailedness level.

        trans: print 2D matrix transposed. default is True.
        -------

        """
        if not issubclass(level.__class__, int):
            raise TypeError('The first argument "level" must be %s not %s %s' %
                            ('int', type(level).__name__, bstr(level, 20)))
        cn = self.__class__.__name__
        if level > 1:
            s = f'{cn}('
            s += 'META: ' + mstr(self._meta, level=level, width=width,
                                 extra=extra,
                                 param_widths=param_widths,
                                 tablefmt=tablefmt, tablefmt1=tablefmt1, tablefmt2=tablefmt2,
                                 excpt=['description'], **kwds)
            s += ' DATA: ' + mstr(self.data, level=level,
                                  excpt=['description'],
                                  tablefmt=tablefmt, tablefmt1=tablefmt1, tablefmt2=tablefmt2,
                                  param_widths=param_widths,
                                  matprint=matprint, trans=trans, heavy=False,
                                  **kwds)
            return s + ')'

        html = 'html' in tablefmt.lower() or 'html' in tablefmt2.lower()
        br = '<br>' if html else '\n'

        from .dataset import make_title_meta_l0
        s, last = make_title_meta_l0(self, level=level, width=width,
                                     tablefmt=tablefmt, tablefmt1=tablefmt1,
                                     tablefmt2=tablefmt2,
                                     extra=extra, center=-1,
                                     param_widths=param_widths,
                                     html=html, excpt=['description'],
                                     **kwds)
        width = len(last)-1
        ds = list(f'"{x}"' for x in self.keys())
        d = 'Total %d Sub-Datasets: %s' % (len(ds), ', '.join(ds))
        if html:
            d = '<center><u>%s</u></center>\n' % d
        else:
            d = d.center(width) + '\n\n'  # + '-' * len(d) + '\n'
        o = ODict(self.data)
        d += o.toString(level=level, heavy=False, center=width,
                        tablefmt=tablefmt, tablefmt1=tablefmt1, tablefmt2=tablefmt2,
                        param_widths=param_widths,
                        matprint=matprint, trans=trans, keyval='SubDataset ',
                        **kwds)
        return '\n\n'.join((x for x in (s, beforedata, d) if len(x))) + '\n' + last

    string = toString
    txt = toString



import yaml
import json
from collections import OrderedDict
# 读取json文件
f = open('VTSE.json', 'r', encoding='utf-8')
# 转化为字典
d = json.loads(f.read(), object_pairs_hook=OrderedDict)
f.close()
import pprint
pprint.pprint(d)
#d = OrderedDict(d)
ym = yaml.dump(d, allow_unicode=True)
# print(ym)
# 转换为yaml
yaml_f = open("VTSE.yml", "w", encoding='utf-8')
yaml_f.write(ym)
yaml_f.close()



# -*- coding: utf-8 -*-

from werkzeug.datastructures import Authorization
from fdi.dataset.testproducts import get_demo_product, get_related_product
from fdi.dataset.classes import Class_Look_Up
from fdi.dataset.deserialize import deserialize
from fdi.pal.poolmanager import PoolManager
from fdi.pal.productstorage import ProductStorage
from fdi.pns.jsonio import getJsonObj
from fdi.utils.common import lls
from fdi.pns.jsonio import auth_headers
from fdi.httppool.model.user import User, getUsers
from fdi.pal.publicclientpool import PublicClientPool

from flask.testing import FlaskClient
from flask import request

from requests.auth import HTTPBasicAuth
import pytest
import importlib
from urllib.error import HTTPError
import requests
import os
import sys
import json
import time
import copy
import getpass
import shlex
import signal
import datetime
import requests
from subprocess import Popen, TimeoutExpired
import logging
import logging.config
from urllib.error import HTTPError, URLError

from logdict import logdict
logging.config.dictConfig(logdict)

logger = logging.getLogger(__name__)
print('**conftest effective logging level** ', logger.getEffectiveLevel())
EX = ' -l /tmp/foo.log'

RUN_SERVER_IN_BACKGROUND = 'python3.8 httppool_app.py --server=httppool_server'
""" set to '' to disable running a pool in the background as the mock. """


TEST_SERVER_LIFE = 600
""" test server time limit in seconds."""

the_session = requests.session()
the_session.secret_key = 'BAD_SECRET_KEY'


@ pytest.fixture(scope='session')
def clean_board():
    importlib.invalidate_caches()
    # importlib.reload(Classes)
    from fdi.dataset.classes import Classes
    return Classes.mapping


@ pytest.fixture(scope="session")
def pc():
    """ get configuration.

    """
    from fdi.utils.getconfig import getConfig as getc
    pns = getc(force=True)
    # logger.debug(json.dumps(pns))

    return pns


@ pytest.fixture(scope="session")
def mock_in_the_background():

    return RUN_SERVER_IN_BACKGROUND


######


SHORT = 'function'


@ pytest.fixture(scope=SHORT)
def new_user_read_write(pc):
    """
    GIVEN a User model
    https://www.patricksoftwareblog.com/testing-a-flask-application-using-pytest/
    """
    new_user = User(pc['username'], pc['password'], roles='read_write')
    headers = auth_headers(pc['username'], pc['password'])
    return new_user, headers


@ pytest.fixture(scope=SHORT)
def new_user_read_only(pc):
    """
    GIVEN a User model
    https://www.patricksoftwareblog.com/testing-a-flask-application-using-pytest/
    """
    users = getUsers(pc)
    new_user = users[pc['ro_user']]
    headers = auth_headers(pc['ro_user'], password=pc['ro_pass'])

    return new_user, headers


@ pytest.fixture(scope=SHORT)
def userpass(pc):
    auth_user = pc['username']
    auth_pass = pc['password']
    return auth_user, auth_pass


@ pytest.fixture(scope="module")
def local_pools_dir(pc):
    """ this is a path in the local OS, where the server runs, used to directly access pool server's internals.

    return: has no trailing '/'
    """
    # http server pool
    schm = 'server'

    # basepath = pc['server_local_pools_dir']
    basepath = PoolManager.PlacePaths[schm]
    # print('WWW ', basepath, pc['api_version'])
    pools_dir = os.path.join(basepath, pc['api_version'])
    return pools_dir

####


@ pytest.fixture(scope="session")
def mock_app(pc, mock_in_the_background):
    if mock_in_the_background:
        yield None
    else:
        from fdi.httppool import create_app
        app = create_app(config_object=pc, level=logger.getEffectiveLevel())
        app.config['TESTING'] = True
        with app.app_context():
            yield app
            # app.


def background_app():
    """ if requied starts a server in the background. """

    # client side.
    # pool url from a local client
    cschm = 'http'
    aburl = cschm + '://' + PoolManager.PlacePaths[cschm]
    pwdir = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))

    pid = os.fork()
    if pid == 0:
        # child process
        # ref https://code.activestate.com/recipes/66012-fork-a-daemon-process-on-unix/
        os.setsid()
        # Redirect standard file descriptors.
        sys.stdin = open('/dev/null', 'r')
        sys.stdout = open('/dev/null', 'w')
        sys.stderr = open('/dev/null', 'w')
        # run server in b/g
        chldlogger = logger
        cmd = shlex.split(RUN_SERVER_IN_BACKGROUND)
        sta = {'command': str(cmd)}
        proc = Popen(cmd, cwd=pwdir, shell=False)
        timeout = TEST_SERVER_LIFE
        try:
            sta['stdout'], sta['stderr'] = proc.communicate(
                timeout=timeout)
        except TimeoutExpired:
            # https://docs.python.org/3.6/library/subprocess.html?highlight=subprocess#subprocess.Popen.communicate
            os.killpg(os.getpgid(proc.pid), signal.SIGTERM)
            msg = 'PID %d is terminated after pre-set timeout %d sec.' % (
                proc.pid, timeout)
        else:
            msg = 'Successful.' if proc.returncode == 0 else 'killed?'
        sta['stdout'], sta['stderr'] = proc.communicate()
        sta['returncode'] = proc.returncode

        sta['message'] = msg
        logg = "Background live server status: %s." % json.dumps(
            dict((k, lls(v, 1000)) for k, v in sta.items()))
        chldlogger.info(logg)
        with open('/tmp/bar.log', 'a') as f:
            f.write(logg)

        assert sta['returncode'] in (
            0, -signal.SIGTERM, -signal.SIGKILL,  signal.SIGHUP), logg
        time.sleep(10000)  # avoid trouble for pytest
        sys.exit(0)
    else:
        # main process
        # wait for checkserver to return 'live'
        time.sleep(2)
        n = 2
        while checkserver(aburl) != 'live':
            logger.debug('No server yet %d' % n)
            n -= 1
            if n == 0:
                break
            time.sleep(2)
        if n:
            msg = 'Made live local server %d' % pid
        else:
            msg = 'Failed running server PID=%s in background. n=%s.' % (
                str(pid), str(n))
        logger.info(msg)
        return pid


def checkserver(aburl):
    """ make sure the server is running when tests start.

    Parameters
    ----------

    Return
    ------
    str
        when aburl points to an live server running either externally to this test (e.g. by `make runpoolserver`), or a server instance by `background_app` fixture created on-demand, server_type is set to 'live'; If no response is returned by `GET`, 'mock' is returned; If server response is abnormal, 'trouble' is returned.
    """

    server_type = None

    # check if data already exists
    try:
        o = getJsonObj(aburl)
        assert o is not None, 'Cannot connect to the server'
        logger.info('Initial server %s response %s' % (aburl, lls(o, 70)))
    except HTTPError as e:
        if e.code == 308:
            logger.info('%s alive. Server response 308' % (aburl))
            server_type = 'live'
        else:
            logger.warning(aburl + ' is alive. but trouble is ')
            logger.warning(e)
            logger.warning('Live server')
            server_type = 'trouble'
    except URLError as e:
        logger.info('Not a live server, because %s' % str(e))
        server_type = 'mock'
    else:
        logger.info('Live server')
        server_type = 'live'
    return server_type

    # assert 'measurements' is not None, 'please start the server to refresh.'
    # initialize test data.


@ pytest.fixture(scope="session")
def live_or_mock(pc, mock_in_the_background, mock_app):
    """ Prepares server absolute base url and common headers for clients to use.

    Based on ``PoolManager.PlacePaths[scheme]`` where ``scheme`` is `http` or `https` and auth info from `pnsconfig` from the configuration file and commandline.

    e.g. ```'http://0.0.0.0:5000/v0.7/', ('foo', 'bar')```

    Return
    ------
    tuple:
       baseurl with no trailing '/' and a string set to 'live' if the
       server os alive, 'mock' if the server is not useable as it is.

    """
    server_type = None

    # client side.
    # pool url from a local client
    cschm = 'http'
    aburl = cschm + '://' + PoolManager.PlacePaths[cschm]
    logger.debug('Check out %s ...' % aburl)
    server_type = checkserver(aburl)
    if server_type == 'mock':
        if mock_in_the_background:
            pid = background_app()
            # None means no mock_in_the_background
            assert pid is not None
            server_type = 'live'
            yield aburl, server_type

            if pid > 0:
                logger.info(
                    'Killing server PID=%d running in the background.' % pid)
                kpg = os.killpg(os.getpgid(pid), signal.SIGTERM)
                logger.info('... killed. rc= %s' % str(kpg))

    elif server_type == 'live':
        yield aburl, server_type
    else:
        logger.error('Invalid server state '+server_type)
        assert 0


@ pytest.fixture(scope=SHORT)
def server(live_or_mock, new_user_read_write):
    """ Server data from r/w user, mock or alive.

    """
    aburl, ty = live_or_mock
    user, headers = new_user_read_write
    headers['server_type'] = ty
    yield aburl, headers


@ pytest.fixture(scope=SHORT)
def server_ro(live_or_mock, new_user_read_only):
    """ Server data from r/w user, alive.

    """
    aburl, ty = live_or_mock
    user, headers = new_user_read_only
    headers['server_type'] = ty
    yield aburl, headers
    del aburl, headers


@ pytest.fixture(scope="module")
def request_context(mock_app):
    """create the app and return the request context as a fixture
       so that this process does not need to be repeated in each test
    https://stackoverflow.com/a/66318710
    """

    yield mock_app.test_request_context


@ pytest.fixture(scope="module")
def client(live_or_mock, mock_app):

    a, server_type = live_or_mock
    if server_type == 'live':
        logger.info('**** requests as client *****')
        with the_session as live_client:
            yield live_client
    elif server_type == 'mock':
        logger.info('**** mock_app as client *****')
        with mock_app.test_client() as client:
            if 0:
                with mock_app.app_context():
                    mock_app.preprocess_request()
            yield client
    else:
        raise ValueError('Invalid server type: ' + server_type)

# @pytest.fixture(scope="module")
# async def a_client(aiohttp_client, server_app, mock_app):
#    if server_app == None:
#        yield aiohttp_client(requests)
#    else:
#        logger.info('**** mock_app as client *****')
#        with mock_app.test_client() as client:
#            with mock_app.app_context():
#                # mock_app.preprocess_request()
#                assert current_app.config["ENV"] == "production"
#            yield aiohttp_client(client)


@ pytest.fixture(scope='module')
def demo_product():
    v = get_demo_product()
    return v, get_related_product()


csdb_pool_id = 'test_csdb_fdi1'


@ pytest.fixture(scope="session")
def urlcsdb(pc):

    return '%s://%s:%d%s/%s' % (pc['scheme'],
                                pc['cloud_host'],
                                pc['cloud_port'],
                                pc['cloud_api_base'],
                                pc['cloud_api_version'])


def make_csdb(poolurl):
    client = the_session
    ps = ProductStorage(poolurl=poolurl, client=client)

    test_pool = ps.getWritablePool(True)  # PublicClientPool(poolurl=url)
    assert test_pool

    return test_pool, poolurl, ps


@ pytest.fixture(scope="session")
def csdb(urlcsdb, pc):
    url = pc['cloud_scheme'] + urlcsdb[len('csdb'):] + '/' + csdb_pool_id
    return make_csdb(url)


@ pytest.fixture(scope=SHORT)
def csdb_new(urlcsdb, pc):
    url = pc['cloud_scheme'] + \
        urlcsdb[len('csdb'):] + '/' + csdb_pool_id + str(int(time.time()))
    # url = pc['cloud_scheme'] + urlcsdb[len('csdb'):] + '/' + csdb_pool_id
    return make_csdb(url)


@ pytest.fixture(scope="function")
def clean_csdb(csdb):
    test_pool, url, ps = csdb

    # url_l = 'http://123.56.102.90:31702/csdb/v1/datatype/list'
    # types = test_pool.client.get(url_l).json()['data']
    # if 'fdi.dataset.testproducts.DemoProduct' in types:
    #     t = len(types)

    ps.wipePool(ignore_error=False)

    # if 'fdi.dataset.testproducts.DemoProduct' in types:
    #     types2 = test_pool.client.get(url_l).json()['data']
    #     assert 'fdi.dataset.testproducts.DemoProduct' in types2

    assert test_pool.isEmpty()
    return test_pool, url, ps


@ pytest.fixture(scope=SHORT)
def tmp_local_storage(tmp_path_factory):
    """ temporary local pool """

    tmppath = tmp_path_factory.mktemp('pools')
    cschm = 'file'
    pdir = str(tmppath.parent)  # PoolManager.PlacePaths[cschm]
    aburl = cschm + '://' + pdir
    poolid = str(tmppath.name)

    pool = PoolManager.getPool(poolid, aburl + '/' + poolid)
    ps = ProductStorage(pool)
    yield ps


@ pytest.fixture(scope=SHORT)
def tmp_remote_storage_no_wipe(server, client, auth):
    """ temporary servered pool with module scope """
    aburl, headers = server
    poolid = 'test_remote_pool'
    pool = PoolManager.getPool(
        poolid, aburl + '/' + poolid, auth=auth, client=client)
    ps = ProductStorage(pool, client=client, auth=auth)
    assert issubclass(ps.getPool(poolid).client.__class__,
                      (requests.Session, FlaskClient))
    yield ps, pool


@ pytest.fixture(scope=SHORT)
def tmp_remote_storage(tmp_remote_storage_no_wipe):
    """ temporary servered pool wiped """
    ps, pool = tmp_remote_storage_no_wipe
    pool.removeAll()
    yield ps


@ pytest.fixture(scope="session")
def tmp_prod_types():
    """ classe of temporary prods with sesion scope """
    ty = []
    se = []
    for n in ('DemoProduct', 'TB', 'TP', 'TC', 'TM', 'SP', 'TCC'):
        cls = Class_Look_Up[n]
        se.append(cls())
        ty.append(cls)
    return ty, se


PRD0_ser = get_demo_product('test-product-0: Demo_Product')
array_ser = Class_Look_Up['ArrayDataset']()


@ pytest.fixture(scope=SHORT)
def tmp_prods(tmp_prod_types):
    """ instances of temporary prods with function scope """
    types, seri = tmp_prod_types
    while True:
        prds = [copy.deepcopy(PRD0_ser)]
        for i, n in enumerate(seri):
            if i == 0:
                continue
            p = copy.deepcopy(n)
            p.description = ('test-product-%d: %s' % (i, n))
            a = copy.deepcopy(array_ser)
            a.data = [[time.time(), n], 's']
            p['the_data'] = a
            prds.append(p)
        logger.debug("Made products: %s" %
                     str(list((p.description, id(p)) for p in prds)))
        res = tuple(prds)
        return res


@ pytest.fixture(scope=SHORT)
def auth(userpass, live_or_mock):

    a, server_type = live_or_mock
    if server_type == 'live':
        return HTTPBasicAuth(*userpass)
    else:
        return Authorization(
            "basic", {"username": userpass[0], "password": userpass[1]})


def gen_pools(url, auth, client, prds):

    tag = str(datetime.datetime.now())
    lst = []
    # n = len(prds)
    n = 2
    for i in range(n):
        poolid = 'test_%d' % i
        poolurl = url + '/' + poolid
        ps = ProductStorage(poolid, poolurl, client=client, auth=auth)
        # the first pool in ps
        pool = ps.getPool(poolid)
        prd = prds[i]
        prd.description = 'lone prod in '+poolid
        ref = ps.save(prd, tag=tag)
        lst.append((pool, prd, ref, tag))
    return lst


@ pytest.fixture(scope=SHORT)
def tmp_pools(server, client, auth, tmp_prods):
    """ generate n pools.

    Return
    ------
    list
        list of tuples containing `ProductPool`, `BaseProduct`, `ProductRef`, `str` for each pool.

"""
    aburl, headers = server
    lst = gen_pools(aburl, auth, client, list(tmp_prods))
    return lst


@ pytest.fixture(scope=SHORT)
def tmp_local_remote_pools(server, client, auth, tmp_prods):
    """ generate n local pools.

    Return
    ------
    list
        list of tuples containing `ProductPool`, `BaseProduct`, `ProductRef`, `str` for each pool.

"""
    aburl, headers = server
    if not aburl.startswith('file://') and not '://127.0.0.1' in aburl and not '://0.0.0.0' in aburl:
        raise ValueError('must be a pool running locally. not %s.' % aburl)
    lst = gen_pools(aburl, auth, client, list(tmp_prods))
    return lst


@ pytest.fixture(scope=SHORT)
def existing_pools(tmp_pools):
    """ return n existing pools.

    Return
    ------
    list
        list of tuples containing `ProductPool`, `BaseProduct`, `ProductRef`, `str` for each pool.

"""
    pools = [p[0] for p in tmp_pools]
    print("get existing pools:", [p.poolname for p in pools])
    return pools



# -*- coding: utf-8 -*-

# from ..utils.common import fullname

import array
import binascii
import gzip
# from .odict import ODict
import logging
import json
import copy
import codecs
import urllib
from collections.abc import Collection, Mapping
from functools import lru_cache
import sys
import datetime
if sys.version_info[0] >= 3:  # + 0.1 * sys.version_info[1] >= 3.3:
    PY3 = True
    strset = (str, bytes)
else:
    PY3 = False
    strset = (str, unicode)

# create logger
logger = logging.getLogger(__name__)
# logger.debug('level %d' %  (logger.getEffectiveLevel()))


class SerializableEncoderAll(json.JSONEncoder):
    """ can encode parameter and product etc such that they can be recovered
    with deserialize().
    Python 3 treats string and unicode as unicode, encoded with utf-8,
    byte blocks as bytes, encoded with utf-8.
    Python 2 treats string as str and unicode as unicode, encoded with utf-8,
    byte blocks as str, encoded with utf-8
    """

    def default(self, obj):
        """
        Parameters
        ----------

        Returns
        -------
        """
        # logger.debug
        # print('&&&& %s %s' % (str(obj.__class__), str(obj)))
        if PY3:
            if issubclass(obj.__class__, bytes):
                return dict(code=codecs.encode(obj, 'hex'), _STID='bytes')
            elif issubclass(obj.__class__, array.array):
                return dict(code=str(binascii.b2a_hex(obj), encoding='ascii'), _STID='array.array_'+obj.typecode)
        if not PY3 and issubclass(obj.__class__, str):
            return dict(code=codec.encode(obj, 'hex'), _STID='bytes')
        if obj is Ellipsis:
            return {'obj': '...', '_STID': 'ellipsis'}
        # print(obj.__getstate__())

        if issubclass(obj.__class__, Serializable):
            return obj.__getstate__()
        print('%%%' + str(obj.__class__))
        return

        # Let the base class default method raise the TypeError
        d = json.JSONEncoder.default(self, obj)
        print('encoded d=' + d)
        return d

    # https://stackoverflow.com/a/63455796/13472124
    base = (str, int, float, bool, type(None))

    def _preprocess(self, obj):
        """ this all only work on the first level of nested objects
        Parameters
        ----------

        Returns
        -------
        """
        oc = obj.__class__
        ocn = type(obj).__name__

        # print('%%%*****prepro ' + ocn)
        # pdb.set_trace()
        # if issubclass(oc, self.base):
        #     # mainly to process string which is a collections (bellow)
        #     return obj
        # elif 0 and issubclass(oc, (Serializable, bytes)):
        #     if issubclass(oc, dict):
        #         # if is both __Getstate__ and Mapping, insert _STID, to a copy
        #         o = copy.copy(obj)
        #         o['_STID'] = obj._STID
        #         return o
        #     return obj
        # elif isinstance(obj, list):
        #     return obj
        # elif issubclass(oc, (Mapping)):
        #     # if all((issubclass(k.__class__, self.base) for k in obj)):
        #     if True:
        #         # JSONEncoder can handle the keys
        #         if isinstance(obj, dict):
        #             return obj
        #         else:
        #             return {'obj': dict(obj), '_STID': ocn}
        #     else:
        #         # This handles the top-level dict keys
        #         return {'obj': [(k, v) for k, v in obj.items()], '_STID': ocn}
        if issubclass(oc, (Collection)):
            return {'obj': list(obj), '_STID': ocn}
        # elif obj is Ellipsis:
        #     return {'obj': '...', '_STID': ocn}

        else:
            return obj

    def iterencode(self, obj, **kwds):
        """
        Parameters
        ----------

        Returns
        -------
        """
        return super().iterencode(self._preprocess(obj), **kwds)


GZIP = False
""" Use ```gzip``` (and ```Bae64``` if needed) to compress. """

SCHEMA = False
""" Output JSONschema instead of JSON erialization. """


class SerializableEncoder(json.JSONEncoder):
    """ can encode parameter and product etc such that they can be recovered
    with deserialize().
    Python 3 treats string and unicode as unicode, encoded with utf-8,
    byte blocks as bytes, encoded with utf-8.
    Python 2 treats string as str and unicode as unicode, encoded with utf-8,
    byte blocks as str, encoded with utf-8
    """

    def default(self, obj):
        """
        Parameters
        ----------

        Returns
        -------
        """
        try:
            # print('%%%' + str(obj.__class__))
            # Let the base class default method raise the TypeError
            d = json.JSONEncoder.default(self, obj)
            # print('d=' + d)
        except TypeError as err:
            try:
                # logger.debug
                # print('&&&& %s %s' % (str(obj.__class__), str(obj)))
                oc = obj.__class__
                if PY3:
                    if issubclass(oc, (datetime.datetime)):
                        if SCHEMA:
                            return '{"$ref": "%s"}' % oc.__name__
                        from ..dataset.finetime import FineTime
                        return dict(
                            code=FineTime.datetimeToFineTime(obj),
                            _STID=oc.__name__+',tai')
                    if issubclass(oc, (bytes, bytearray)):
                        if SCHEMA:
                            return '{"$ref": "bytes"}'
                        if GZIP:
                            r = dict(code=binascii.b2a_base64(
                                gzip.compress(obj, 5)).decode('ascii'),
                                _STID=oc.__name__ + ',gz,b64')
                        else:
                            r = dict(code=obj.hex(), _STID=oc.__name__)
                        return r
                    elif issubclass(oc, array.array):
                        if SCHEMA:
                            return '{"$ref": "%s"}' % oc.__name__
                        if GZIP:
                            r = dict(code=binascii.b2a_base64(
                                gzip.compress(obj, 5)).decode('ascii'),
                                _STID='a.array_%s,gz,b64' % obj.typecode)
                        else:
                            r = dict(code=str(codecs.encode(obj, 'hex'),
                                              encoding='ascii'),
                                     _STID='a.array_'+obj.typecode)
                        return r
                if not PY3 and issubclass(oc, str):
                    # return dict(code=codec.encode(obj, 'hex'), _STID='bytes')
                    assert False, lls(obj, 50)
                    if GZIP:
                        if SCHEMA:
                            return '{"$ref": "%s"}' % 'bytes'
                        return dict(code=gzip.compress(obj, 5),
                                    _STID='bytes,gz')
                    else:
                        return obj
                if obj is Ellipsis:
                    if SCHEMA:
                        return '{"$ref": "%s"}' % oc.__name__
                    return {'obj': '...', '_STID': 'ellipsis'}
                if issubclass(oc, type):
                    if SCHEMA:
                        return '{"$ref": "%s"}' % oc.__name__
                    return {'obj': obj.__name__, '_STID': 'dtype'}
                if hasattr(obj, 'serializable'):
                    if SCHEMA:
                        return '{%s}' % obj.schema()
                    try:
                        typ = obj.type
                    except (LookupError, AttributeError):
                        typ = None
                    if typ in ['image/svg']:
                        ser = obj.serializable()[ATTR+'data']
                        ser = urllib.parse.quote(ser)
                    # print(obj.serializable())
                    return obj.serializable()
                try:
                    return dict(obj)
                except Exception:
                    return list(obj)
            except Exception as e:
                print('Serialization failed.' + str(e))
                raise


#    obj = json.loads(jstring)

def serialize(o, cls=None, **kwds):
    """ return JSON using special encoder SerializableEncoder

    Parameterts
    -----------

    Returns
    -------
    """
    if not cls:
        cls = SerializableEncoder
    return json.dumps(o, cls=cls, allow_nan=True, **kwds)


@lru_cache(maxsize=256)
def get_schema_with_classname(cls_name, store=None):
    if store is None:
        store = makeSchemeStore()
    for sch in store:
        if cls_name == 'array':
            n = 'a_array'
        else:
            n = cls_name
        if sch.endswith('/%s' % n):
            return sch, store[sch]
    # did not find.
    return None, None


ATTR = '_ATTR_'
LEN_ATTR = len(ATTR)


class Serializable():
    """ mh: Can be serialized.
    Has a _STID  instance property to show its class information. """

    def __init__(self, *args, **kwds):
        """

        Parameters
        ----------

        Returns
        -------
        """
        super().__init__(*args, **kwds)
        sc = self.__class__
        # print('@@@ ' + sc.__name__, str(issubclass(sc, dict)))

        self._STID = sc.__name__

    def serialized(self, indent=None):
        """
        Parameters
        ----------


        Returns
        -------
        """
        return serialize(self, indent=indent)

    def __repr__(self):

        co = ', '.join(str(k)+'=' + ('"'+v+'"'
                                     if issubclass(v.__class__, str)
                                     else str(v))
                       for k, v in self.__getstate__().items()
                       )
        return self.__class__.__name__ + '(' + co + ')'

    def __getstate__(self):
        """ returns an ordered ddict that has all state info of this object.
        Subclasses should override this function.
        Parameters
        ----------

        Returns
        -------
        """
        raise NotImplementedError()

    def __setstate__(self, state):
        """
        Parameters
        ----------

        Returns
        -------
        """
        for name in state.keys():
            if name.startswith(ATTR):
                k2 = name[LEN_ATTR:]
                self.__setattr__(k2, state[name])
            elif name == '_STID':
                pass
            elif hasattr(self, '__setitem__'):
                self[name] = state[name]
            else:
                self.__setattr__(name, state[name])

    def __reduce_ex__(self, protocol):
        """
        Parameters
        ----------

        Returns
        -------
        """
        def func(): return self.__class__()
        args = tuple()
        state = self.__getstate__()
        return func, args, state

    def __reduce__(self):
        """
        Parameters
        ----------

        Returns
        -------
        """
        return self.__reduce_ex__(4)

    def serializable(self):
        """ Can be encoded with serializableEncoder.

        Return
        ------
        dict
             The state variables plus the Serialization Type ID with ```_STID``` as uts key.
        """
        s = copy.copy(self.__getstate__())
        # make sure _STID is the last, for pools to ID data.
        if '_STID' in s:
            del s['_STID']
        s.update({'_STID': self._STID})
        return s

    def schema(self):
        """ Get schema definition using the FDI standard schema set in `FDI_SCHEMA_STORE`. Subclassing to add more schemas.
        """
        sid, sch = get_schema_with_classname(self.__class__.__name__)
        return sch

    def yaml(self, *args, **kwds):
        """ Get a YAML representation. """
        from ..utils.ydump import ydump, yinit
        yinit()
        return ydump(self, *args, **kwds)

    def tree(self, *args, **kwds):
        """ Get a directory-tree-like representation. """
        from ..utils.tree import tree

        return '\n'.join(tree(self, *args, **kwds))

    def fits(self, *args, **kwds):
        """ Get a FITS representation. """
        from ..utils.tofits import toFits, FITS_INSTALLED

        if not FITS_INSTALLED:
            raise NotImplemented(
                'Astropy not installed. Include SCI in extra-dependency when installing FDI.')

        return toFits(self, *args, **kwds)

    def html(self, extra=False, param_widths=-1, **kwds):
        """ Get a HTML representation. """

        return self.toString(level=0,
                             tablefmt='unsafehtml',
                             tablefmt1='unsafehtml',
                             tablefmt2='unsafehtml',
                             extra=extra,
                             param_widths=param_widths, **kwds)

    def jsonPath(self, expr, val='simple', sep='/', indent=None, *args, **kwds):
        from ..utils.jsonpath import jsonPath

        return jsonPath(self.data, expr=expr, val=val, indent=indent, *args, **kwds)

    def fetch(self, paths, exe=['is'], not_quoted=True):
        from ..utils.fetch import fetch

        return fetch(paths, self, re='', exe=exe, not_quoted=not_quoted)



# -*- coding: utf-8 -*-

from ..dataset.deserialize import deserialize_args

from operator import methodcaller
import inspect
from itertools import chain


def fetch(paths, nested, re='', sep='/', exe=['is'], not_quoted=True):
    """ Use paths to access values of internal elements of a nested python object.

    :paths: 1). If given as a string, the string will be splitted with `sep` into a list of strings, then go on to 2); 2). If given as a list of strings, its 0th member is to match one of the first level of nested attributes, keys, or method names. If the list was made in 1) the 0th member will be converted to an integer if possible.
    * if the 0th member is a string and can be parsed by :meth:`deserialize_args`, the result to used as te named method and its arguments.
    * if that fails, it will be taken as a string and check if there is a match in keys (members);
    * else search in attributes.

    :nested: a live nested data structure.
    :re: datapath representation for `nested`. Can be applied to reproduce the result.
    :exe: 1) A list of patterns which if found in the name of a method/function the matching method/function is allowed to run. 2) If one of the pattern is '*', all methods/functions are allowed to run. 3) If a pattern starts with a '-' then the matching method/function to the pattern ('-' removed) is not allowed to run (overriding previous rules.
    :not_quoted: the method-args string is not encoded with `quote`.

    Result
    A found object and a string of python code e.g.  '.meta["speed"].isValid()'
    """
    if issubclass(paths.__class__, str):
        # treat integers in data path as string
        paths = paths.strip(' ').strip(sep).split(sep)
        from_str = True
    else:
        from_str = False

    if len(paths) == 0 or paths[0] == '':
        return nested, re

    p0 = paths[0]
    found_method = None
    # print('>>>> ', p0)

    if from_str:
        try:
            p0 = int(p0)
        except (ValueError, TypeError):
            # nested expects an integer
            pass
        # f'{p0} cannot be converted to an integer.').with_traceback(sys.exc_info()[2])
    try:
        v0 = nested[p0]
        q = '"' if issubclass(p0.__class__, str) else ''
        rep = re + '['+q + str(p0) + q + ']'
        if len(paths) == 1:
            return v0, rep
        return fetch(paths[1:], v0, re=rep, sep=sep, exe=exe)
    except (TypeError, KeyError, ValueError):
        pass

    if not issubclass(p0.__class__, str):
        raise TypeError(f'{p0} should be a string, not {p0.__class__}.')

    # get command positional arguments and keyword arguments
    code, m_args, kwds = deserialize_args(
        all_args=p0, not_quoted=not_quoted)
    p0, args = m_args[0], m_args[1:]

    if hasattr(nested, p0):
        v0 = getattr(nested, p0)
        rep = re + '.' + p0
        if inspect.ismethod(v0) or inspect.isfunction(v0):
            if '*' in exe:
                can_exec = True
            else:   # TODO test
                can_exec = any(patt in p0 for patt in exe if patt[0] != '-')
            can_exec = not any(
                patt[1:] in p0 for patt in exe if patt[0] == '-')
            if can_exec:
                # assemble expression of keywords args from deserialize_args
                kwdsexpr = [str(k)+'='+str(v) for k, v in kwds.items()]
                # assemble positional and keywords args
                all_args_expr = ', '.join(chain(map(str, args), kwdsexpr))
                # return execution results and path-representation
                # return f'{rep}({all_args_expr})', f'{rep}({all_args_expr})'

                # v0(*[], **{}) is not v0() !
                res = v0(*args, **kwds)
                v0 = res
                rep = f'{rep}({all_args_expr})'
        # not executable
        if len(paths) == 1:
            return v0, rep
        return fetch(paths[1:], v0, re=rep, sep=sep, exe=exe)
    # elif issubclass(nested.__class__, Squence):
    #     try:
    #         i = int(p0)
    #     except ValueError:
    #         pass
    #     else:
    #         return fetch(paths[1:], nested[i], re=rep, sep=sep, exe=exe)
    # not methods, attribute or member
    # if found_method:
        # return methodcaller(p0)(nested), rep + '()'
    #    return found_method(), rep + '()'
    return None, '%s has no attribute or member: %s.' % (re, p0)
